# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Hashable, Mapping, Sequence
from typing import Any, Dict, Optional, Tuple, Union

import dask.array as da
import numpy as np
import pandas as pd
import xarray as xr

from xcube.core.verify import assert_cube

DEFAULT_INDEX_NAME_PATTERN = "{name}_index"
DEFAULT_REF_NAME_PATTERN = "{name}_ref"
INDEX_DIM_NAME = "idx"

POINT_INTERP_METHOD_NEAREST = "nearest"
POINT_INTERP_METHOD_LINEAR = "linear"
DEFAULT_INTERP_POINT_METHOD = POINT_INTERP_METHOD_NEAREST

SeriesLike = Union[
    np.ndarray, da.Array, xr.DataArray, pd.Series, Sequence[Union[int, float]]
]

PointsLike = Union[xr.Dataset, pd.DataFrame, Mapping[str, SeriesLike]]


def get_cube_values_for_points(
    cube: xr.Dataset,
    points: PointsLike,
    var_names: Sequence[str] = None,
    include_coords: bool = False,
    include_bounds: bool = False,
    include_indexes: bool = False,
    index_name_pattern: str = DEFAULT_INDEX_NAME_PATTERN,
    include_refs: bool = False,
    ref_name_pattern: str = DEFAULT_REF_NAME_PATTERN,
    method: str = DEFAULT_INTERP_POINT_METHOD,
    cube_asserted: bool = False,
) -> xr.Dataset:
    """Extract values from *cube* variables at given
    coordinates in *points*.

    Returns a new dataset with values of variables from *cube* selected
    by the coordinate columns
    provided in *points*. All variables will be 1-D and have the same order
    as the rows in points.

    Args:
        cube: The cube dataset.
        points: Dictionary that maps dimension name to coordinate
            arrays.
        var_names: An optional list of names of data variables in *cube*
            whose values shall be extracted.
        include_coords: Whether to include the cube coordinates for each
            point in return value.
        include_bounds: Whether to include the cube coordinate
            boundaries (if any) for each point in return value.
        include_indexes: Whether to include computed indexes into the
            cube for each point in return value.
        index_name_pattern: A naming pattern for the computed index
            columns. Must include "{name}" which will be replaced by the
            index' dimension name.
        include_refs: Whether to include point (reference) values from
            *points* in return value.
        ref_name_pattern: A naming pattern for the computed point data
            columns. Must include "{name}" which will be replaced by the
            point's attribute name.
        method: "nearest" or "linear".
        cube_asserted: If False, *cube* will be verified, otherwise it
            is expected to be a valid cube.

    Returns:
        A new data frame whose columns are values from *cube* variables
        at given *points*.
    """
    if not cube_asserted:
        assert_cube(cube)

    index_dtype = np.int64 if method == POINT_INTERP_METHOD_NEAREST else np.float64

    point_indexes = get_cube_point_indexes(
        cube,
        points,
        index_name_pattern=index_name_pattern,
        index_dtype=index_dtype,
        cube_asserted=True,
    )

    cube_values = get_cube_values_for_indexes(
        cube,
        point_indexes,
        include_coords,
        include_bounds,
        data_var_names=var_names,
        index_name_pattern=index_name_pattern,
        method=method,
        cube_asserted=True,
    )

    if include_indexes:
        cube_values.update(point_indexes)

    if include_refs:
        point_refs = xr.Dataset(
            {
                ref_name_pattern.format(name=name): xr.DataArray(
                    points[name], dims=[INDEX_DIM_NAME]
                )
                for name in points.keys()
            }
        )

        cube_values.update(point_refs)

    # The (coordinate) variable named INDEX_DIM_NAME should not exist,
    # otherwise cube_values.to_dataframe() will use an empty string
    # to name that series "" in the output DataFrame instance.
    # This seems to be an issue of xarray 0.18.2+.
    if INDEX_DIM_NAME in cube_values:
        cube_values = cube_values.drop_vars(INDEX_DIM_NAME)

    return cube_values


def get_cube_values_for_indexes(
    cube: xr.Dataset,
    indexes: Union[xr.Dataset, pd.DataFrame, Mapping[str, Any]],
    include_coords: bool = False,
    include_bounds: bool = False,
    data_var_names: Sequence[str] = None,
    index_name_pattern: str = DEFAULT_INDEX_NAME_PATTERN,
    method: str = DEFAULT_INTERP_POINT_METHOD,
    cube_asserted: bool = False,
) -> xr.Dataset:
    """Get values from the *cube* at given *indexes*.

    Args:
        cube: A cube dataset.
        indexes: A mapping from column names to index and fraction
            arrays for all cube dimensions.
        include_coords: Whether to include the cube coordinates for each
            point in return value.
        include_bounds: Whether to include the cube coordinate
            boundaries (if any) for each point in return value.
        data_var_names: An optional list of names of data variables in
            *cube* whose values shall be extracted.
        index_name_pattern: A naming pattern for the computed indexes
            columns. Must include "{name}" which will be replaced by the
            dimension name.
        method: "nearest" or "linear".
        cube_asserted: If False, *cube* will be verified, otherwise it
            is expected to be a valid cube.

    Returns:
        A new data frame whose columns are values from *cube* variables
        at given *indexes*.
    """
    if not cube_asserted:
        assert_cube(cube)

    if method not in {POINT_INTERP_METHOD_NEAREST, POINT_INTERP_METHOD_LINEAR}:
        raise ValueError(f"invalid method {method!r}")
    if method != POINT_INTERP_METHOD_NEAREST:
        raise NotImplementedError(f"method {method!r} not yet implemented")

    all_data_var_names = tuple(cube.data_vars.keys())
    if len(all_data_var_names) == 0:
        raise ValueError("cube is empty")

    if data_var_names is not None:
        if len(data_var_names) == 0:
            return _empty_dataset_from_points(indexes)
        for var_name in data_var_names:
            if var_name not in cube.data_vars:
                raise ValueError(f"variable {var_name!r} not found in cube")
    else:
        data_var_names = all_data_var_names

    dim_names = cube[data_var_names[0]].dims
    num_dims = len(dim_names)
    index_names = [index_name_pattern.format(name=dim_name) for dim_name in dim_names]
    indexes, num_points = _normalize_series(
        indexes, index_names, force_dataset=True, param_name="indexes"
    )
    if num_points == 0:
        return _empty_dataset_from_points(indexes)
    cube = xr.Dataset(
        {var_name: cube[var_name] for var_name in data_var_names}, coords=cube.coords
    )

    new_bounds_vars = {}
    bounds_var_names = _get_coord_bounds_var_names(cube)
    drop_coords = None
    if bounds_var_names:
        if include_bounds:
            # Flatten any coordinate bounds variables
            for var_name, bnds_var_name in bounds_var_names.items():
                bnds_var = cube[bnds_var_name]
                new_bounds_vars[f"{var_name}_lower"] = bnds_var[:, 0]
                new_bounds_vars[f"{var_name}_upper"] = bnds_var[:, 1]
            cube = cube.assign_coords(**new_bounds_vars)
        cube = cube.drop_vars(bounds_var_names.values())
        if not include_coords:
            drop_coords = set(cube.coords).difference(new_bounds_vars.keys())
    else:
        if not include_coords:
            drop_coords = set(cube.coords)

    # Generate a validation condition so we can
    # filter out invalid rows (where any index == -1)
    is_valid_point = None
    for index_name in index_names:
        col = indexes[index_name]
        condition = col >= 0 if np.issubdtype(col.dtype, np.integer) else np.isnan(col)
        if is_valid_point is None:
            is_valid_point = condition
        else:
            is_valid_point = np.logical_and(is_valid_point, condition)

    num_valid_points = np.count_nonzero(is_valid_point)
    if num_valid_points == num_points:
        # All indexes valid
        cube_selector = {dim_names[i]: indexes[index_names[i]] for i in range(num_dims)}
        cube_values = cube.isel(cube_selector)
    elif num_valid_points == 0:
        # All indexes are invalid
        new_bounds_vars = {}
        for var_name in cube.variables:
            new_bounds_vars[var_name] = _empty_points_var(cube[var_name], num_points)
        cube_values = xr.Dataset(new_bounds_vars)
    else:
        # Some invalid indexes
        idx = np.arange(num_points)
        good_idx = idx[is_valid_point.values]
        idx_dim_name = indexes[index_names[0]].dims[0]
        good_indexes = indexes.isel({idx_dim_name: good_idx})

        cube_selector = {
            dim_names[i]: good_indexes[index_names[i]] for i in range(num_dims)
        }
        cube_values = cube.isel(cube_selector)

        new_bounds_vars = {}
        for var_name in cube.variables:
            var = cube_values[var_name]
            new_var = _empty_points_var(var, num_points)
            new_var[good_idx] = var
            new_bounds_vars[var_name] = new_var

        cube_values = xr.Dataset(new_bounds_vars)

    if drop_coords:
        cube_values = cube_values.drop_vars(drop_coords)

    return cube_values


def get_cube_point_indexes(
    cube: xr.Dataset,
    points: PointsLike,
    dim_name_mapping: Mapping[str, str] = None,
    index_name_pattern: str = DEFAULT_INDEX_NAME_PATTERN,
    index_dtype=np.float64,
    cube_asserted: bool = False,
) -> xr.Dataset:
    """Get indexes of given point coordinates *points* into the given *dataset*.

    Args:
        cube: The cube dataset.
        points: A mapping from column names to column data arrays, which
            must all have the same length.
        dim_name_mapping: A mapping from dimension names in *cube* to
            column names in *points*.
        index_name_pattern: A naming pattern for the computed indexes
            columns. Must include "{name}" which will be replaced by the
            dimension name.
        index_dtype: Numpy data type for the indexes. If it is a
            floating point type (default), then *indexes* will contain
            fractions, which may be used for interpolation. For out-of-
            range coordinates in *points*, indexes will be -1 if
            *index_dtype* is an integer type, and NaN, if *index_dtype*
            is a floating point types.
        cube_asserted: If False, *cube* will be verified, otherwise it
            is expected to be a valid cube.

    Returns:
        A dataset containing the index columns.
    """
    if not cube_asserted:
        assert_cube(cube)

    dim_name_mapping = dim_name_mapping if dim_name_mapping is not None else {}
    dim_names = _get_cube_data_var_dims(cube)
    col_names = [
        dim_name_mapping.get(str(dim_name), dim_name) for dim_name in dim_names
    ]

    points, _ = _normalize_series(
        points, col_names, force_dataset=False, param_name="points"
    )

    indexes = []
    for dim_name, col_name in zip(dim_names, col_names):
        col = points[col_name]
        coord_indexes = get_dataset_indexes(
            cube, str(dim_name), col, index_dtype=index_dtype
        )
        indexes.append(
            (
                index_name_pattern.format(name=dim_name),
                xr.DataArray(coord_indexes, dims=[INDEX_DIM_NAME]),
            )
        )

    return xr.Dataset(dict(indexes))


def get_dataset_indexes(
    dataset: xr.Dataset,
    coord_var_name: str,
    coord_values: SeriesLike,
    index_dtype=np.float64,
) -> Union[xr.DataArray, np.ndarray]:
    """Compute the indexes and their fractions into a coordinate variable
    *coord_var_name* of a *dataset* for the given coordinate values
    *coord_values*.

    The coordinate variable's labels must be monotonic increasing or
    decreasing, otherwise the result will be nonsense.

    For any value in *coord_values* that is out of the bounds of the
    coordinate variable's values, the index depends on the value of
    *index_dtype*. If *index_dtype* is an integer type, invalid indexes
    are encoded as -1 while for floating point types, NaN will be used.

    Returns a tuple of indexes as int64 array and fractions as
    float64 array.

    Args:
        dataset: A cube dataset.
        coord_var_name: Name of a coordinate variable.
        coord_values: Array-like coordinate values.
        index_dtype: Numpy data type for the indexes. If it is floating
            point type (default), then *indexes* contain fractions,
            which may be used for interpolation. If *dtype* is an
            integer type out-of-range coordinates are indicated by index
            -1, and NaN if it is is a floating point type.

    Returns:
        The indexes and their fractions as a tuple of numpy int64 and
        float64 arrays.
    """
    coord_var = dataset[coord_var_name]
    n1 = coord_var.size
    n2 = n1 + 1

    coord_bounds_var = _get_bounds_var(dataset, coord_var_name)
    if coord_bounds_var is not None:
        coord_bounds = coord_bounds_var.values
        if np.issubdtype(coord_bounds.dtype, np.datetime64):
            coord_bounds = coord_bounds.astype(np.uint64)
        is_inverse = (coord_bounds[0, 1] - coord_bounds[0, 0]) < 0
        if is_inverse:
            coord_bounds = coord_bounds[::-1, ::-1]
        coords = np.zeros(n2, dtype=coord_bounds.dtype)
        coords[0:-1] = coord_bounds[:, 0]
        coords[-1] = coord_bounds[-1, 1]
    elif coord_var.size > 1:
        center_coords = coord_var.values
        if np.issubdtype(center_coords.dtype, np.datetime64):
            center_coords = center_coords.astype(np.uint64)
        is_inverse = (center_coords[-1] - center_coords[0]) < 0
        if is_inverse:
            center_coords = center_coords[::-1]
        deltas = np.zeros(n2, dtype=center_coords.dtype)
        deltas[0:-2] = np.diff(center_coords)
        deltas[-2] = deltas[-3]
        deltas[-1] = deltas[-3]
        coords = np.zeros(n2, dtype=center_coords.dtype)
        coords[0:-1] = center_coords
        coords[-1] = coords[-2] + deltas[-1]
        if np.issubdtype(deltas.dtype, np.integer):
            coords -= deltas // 2
        else:
            coords -= 0.5 * deltas
    else:
        raise ValueError(
            f"cannot determine cell boundaries for"
            f" coordinate variable {coord_var_name!r}"
            f" of size {coord_var.size}"
        )

    if np.issubdtype(coord_values.dtype, np.datetime64):
        try:
            coord_values = coord_values.astype(np.uint64)
        except TypeError:
            # Fixes https://github.com/dcs4cop/xcube/issues/95
            coord_values = coord_values.values.astype(np.uint64)

    indexes = np.linspace(0.0, n1, n2, dtype=np.float64)
    interp_indexes = np.interp(coord_values, coords, indexes, left=-1, right=-1)
    if is_inverse:
        i = interp_indexes >= 0
        interp_indexes[i] = n1 - interp_indexes[i]
    interp_indexes[interp_indexes >= n1] = n1 - 1e-9
    if np.issubdtype(index_dtype, np.integer):
        interp_indexes = interp_indexes.astype(index_dtype)
    else:
        interp_indexes[interp_indexes < 0] = np.nan

    return interp_indexes


def _empty_points_var(var: xr.DataArray, num_points: int):
    fill_value = 0 if np.issubdtype(var.dtype, np.integer) else np.nan
    return xr.DataArray(
        np.full(num_points, fill_value, dtype=var.dtype),
        dims=[INDEX_DIM_NAME],
        attrs=var.attrs,
    )


def _normalize_series(
    points: PointsLike,
    col_names: Sequence[str],
    force_dataset: bool = False,
    param_name: str = "points",
) -> tuple[PointsLike, int]:
    if not isinstance(points, (xr.Dataset, pd.DataFrame)):
        new_points = {}
        for col_name in col_names:
            if col_name in points:
                col = points[col_name]
                if not isinstance(col, xr.DataArray):
                    col = xr.DataArray(col)
                new_points[col_name] = col
        points = new_points

    num_points = None
    for col_name in col_names:
        if col_name not in points:
            raise ValueError(f"column {col_name!r} not found in {param_name}")
        col = points[col_name]
        if len(col.shape) != 1:
            raise ValueError(
                f"column {col_name!r} in {param_name} must be"
                f" one-dimensional, but has shape {col.shape!r}"
            )
        if num_points is None:
            num_points = len(col)
        elif num_points != len(col):
            raise ValueError(
                f"column sizes in {param_name} must be all the"
                f" same, but found {len(points[col_names[0]])}"
                f" for column {col_names[0]!r} and {num_points}"
                f" for column {col_name!r}"
            )

    if num_points is None:
        raise ValueError(f"{param_name} has no valid columns")

    if force_dataset:
        if isinstance(points, pd.DataFrame):
            points = xr.Dataset.from_dataframe(points)
        elif not isinstance(points, xr.DataArray):
            points = xr.Dataset(points)

    return points, num_points


def _get_coord_bounds_var_names(dataset: xr.Dataset) -> dict[Hashable, Any]:
    bnds_var_names = {}
    for var_name in dataset.coords:
        var = dataset[var_name]
        bnds_var_name = var.attrs.get("bounds", f"{var_name}_bnds")
        if bnds_var_name not in bnds_var_names and bnds_var_name in dataset:
            bnds_var = dataset[bnds_var_name]
            if bnds_var.shape == (var.shape[0], 2):
                bnds_var_names[var_name] = bnds_var_name
    return bnds_var_names


def _get_cube_data_var_dims(cube: xr.Dataset) -> tuple[Hashable, ...]:
    for var in cube.data_vars.values():
        return var.dims
    raise ValueError("cube dataset is empty")


def _get_bounds_var(dataset: xr.Dataset, var_name: str) -> Optional[xr.DataArray]:
    var = dataset[var_name]
    if len(var.shape) == 1:
        bounds_var_name = var.attrs.get("bounds", f"{var_name}_bnds")
        if bounds_var_name in dataset:
            bounds_var = dataset[bounds_var_name]
            if bounds_var.dtype == var.dtype and bounds_var.shape == (var.size, 2):
                return bounds_var
    return None


def _empty_dataset_from_points(data: Any) -> xr.Dataset:
    return xr.Dataset(coords=data.coords if hasattr(data, "coords") else None)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import asyncio
import json
import socket
import threading
import time
import unittest
from abc import ABC
from contextlib import closing
from typing import IO, Any, Iterable

import urllib3
from urllib3.exceptions import MaxRetryError

from xcube.server.server import Server
from xcube.server.webservers.tornado import TornadoFramework
from xcube.util.extension import ExtensionRegistry

Body = dict | str | bytes | IO[Any] | Iterable[bytes]
JsonData = None | bool | int | float | str | list | dict


# taken from https://stackoverflow.com/a/45690594
def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("localhost", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


class ServerTestCase(unittest.TestCase, ABC):
    """Base test class for unit tests of server framework

    This class is in the main xcube codebase rather than among the unit
    tests because xcube server is a framework module and API extension
    developers should be able to make use of the testing module.
    """

    def setUp(self) -> None:
        self.port = find_free_port()
        config = self.get_config()
        config.update({"port": self.port, "address": "localhost"})
        self.add_config(config)
        asyncio.set_event_loop(asyncio.new_event_loop())
        er = self.get_extension_registry()
        self.add_extension(er)
        self.server = Server(
            framework=TornadoFramework(), config=config, extension_registry=er
        )
        tornado = threading.Thread(target=self.server.start)
        tornado.daemon = True
        tornado.start()

        self.http = urllib3.PoolManager()

        # Taking care that server is fully started before tests make requests.
        # Fixing https://github.com/dcs4cop/xcube/issues/899
        while True:
            try:
                self.fetch("/")
                break
            except (MaxRetryError, TimeoutError):
                time.sleep(0.1)

    def tearDown(self) -> None:
        self.server.stop()

    # noinspection PyMethodMayBeStatic
    def get_extension_registry(self) -> ExtensionRegistry:
        return ExtensionRegistry()

    def add_extension(self, er: ExtensionRegistry) -> None:
        pass

    # noinspection PyMethodMayBeStatic
    def get_config(self) -> dict[str, Any]:
        return {}

    def add_config(self, config: dict) -> None:
        pass

    def fetch(
        self,
        path: str,
        method: str = "GET",
        headers: dict[str, str] | None = None,
        body: Body | None = None,
    ) -> urllib3.response.HTTPResponse:
        """Fetches the response for given request.

        Args:
            path: Route request path
            method: HTTP request method, default to "GET"
            headers: HTTP request headers
            body: HTTP request body

        Returns:
            A response object
        """
        if path.startswith("/"):
            path = path[1:]
        url = f"http://localhost:{self.port}/{path}"
        if isinstance(body, dict):
            body = bytes(json.dumps(body), "utf-8")
        return self.http.request(method, url, headers=headers, body=body)

    def fetch_json(
        self,
        path: str,
        method: str = "GET",
        headers: dict[str, str] | None = None,
        body: Body | None = None,
    ) -> tuple[JsonData, int]:
        """Fetches the response's JSON value for given request.

        Args:
            path: Route request path
            method: HTTP request method, default to "GET"
            headers: HTTP request headers
            body: HTTP request body

        Returns:
            A json object parsed from response data
        """
        response = self.fetch(path, method=method, headers=headers, body=body)
        return json.loads(response.data), response.status

    def assertResponseOK(self, response: urllib3.response.HTTPResponse):
        """Assert that response has status code 200.

        Args:
            response: The response from a ``self.fetch()`` call
        """
        self.assertResponse(response, expected_status=200)

    def assertResourceNotFoundResponse(
        self, response: urllib3.response.HTTPResponse, expected_message: str = None
    ):
        """Assert that response
        has the given status code 404, and
        has the given error message *expected_message*, if given.

        Args:
            response: The response from a ``self.fetch()`` call
            expected_message: The expected error message or part of it.
        """
        self.assertResponse(
            response, expected_status=404, expected_message=expected_message
        )

    def assertBadRequestResponse(
        self, response: urllib3.response.HTTPResponse, expected_message: str = None
    ):
        """Assert that response
        has the given status code 400, and
        has the given error message *expected_message*, if given.

        Args:
            response: The response from a ``self.fetch()`` call
            expected_message: The expected error message or part of it.
        """
        self.assertResponse(
            response, expected_status=400, expected_message=expected_message
        )

    def assertResponse(
        self,
        response: urllib3.response.HTTPResponse,
        expected_status: int | None = None,
        expected_message: str | None = None,
    ):
        """Assert that response
        has the given status code *expected_status*, if given, and
        has the given error message *expected_message*, if given.

        Args:
            response: The response from a ``self.fetch()`` call
            expected_status: The expected status code.
            expected_message: The expected error message or part of it.
        """
        message = self.parse_error_message(response)
        if expected_status is not None:
            self.assertEqual(
                expected_status, response.status, msg=message or response.reason
            )
        if expected_message is not None:
            self.assertIsNotNone(message, msg="error message not available")
            self.assertIsInstance(
                message, str, msg=f"error message must be a string, was {message!r}"
            )
            self.assertIn(expected_message, message)

    @classmethod
    def parse_error_message(cls, response: urllib3.response.HTTPResponse) -> str | None:
        """Parse an error message from the given response object."""
        # noinspection PyBroadException
        try:
            result = json.loads(response.data)
            return result["error"]["message"]
        except BaseException:
            return None


# TODO (forman): Remove before final PR
ServerTest = ServerTestCase



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Callable

import xarray as xr

from xcube.util.assertions import assert_instance


class ExprVar:
    """A wrapped `xarray.DataArray` to allow safe access in expressions.

    Args:
        da: The `xarray.DataArray` to be wrapped.
    """

    def __init__(self, da: xr.DataArray):
        assert_instance(da, xr.DataArray, name="da")
        # Note that the double underscore protects access by "name mangling"
        self.__da = da

    ################################################
    # Binary operations - comparisons

    def __eq__(self, other):
        return self.__wrap(self.__da == self.__unwrap(other))

    def __ne__(self, other):
        return self.__wrap(self.__da != self.__unwrap(other))

    def __le__(self, other):
        return self.__wrap(self.__da <= self.__unwrap(other))

    def __lt__(self, other):
        return self.__wrap(self.__da < self.__unwrap(other))

    def __ge__(self, other):
        return self.__wrap(self.__da >= self.__unwrap(other))

    def __gt__(self, other):
        return self.__wrap(self.__da > self.__unwrap(other))

    ################################################
    # Binary operations - emulating numeric type

    def __add__(self, other):
        return self.__wrap(self.__da + self.__unwrap(other))

    def __radd__(self, other):
        return self.__wrap(self.__unwrap(other) + self.__da)

    def __sub__(self, other):
        return self.__wrap(self.__da - self.__unwrap(other))

    def __rsub__(self, other):
        return self.__wrap(self.__unwrap(other) - self.__da)

    def __mul__(self, other):
        return self.__wrap(self.__da * self.__unwrap(other))

    def __rmul__(self, other):
        return self.__wrap(self.__unwrap(other) * self.__da)

    def __truediv__(self, other):
        return self.__wrap(self.__da / self.__unwrap(other))

    def __rtruediv__(self, other):
        return self.__wrap(self.__unwrap(other) / self.__da)

    def __floordiv__(self, other):
        return self.__wrap(self.__da // self.__unwrap(other))

    def __rfloordiv__(self, other):
        return self.__wrap(self.__unwrap(other) // self.__da)

    def __mod__(self, other):
        return self.__wrap(self.__da % self.__unwrap(other))

    def __rmod__(self, other):
        return self.__wrap(self.__unwrap(other) % self.__da)

    def __pow__(self, power):
        return self.__wrap(self.__da ** self.__unwrap(power))

    def __rpow__(self, other):
        return self.__wrap(self.__unwrap(other) ** self.__da)

    def __lshift__(self, other):
        return self.__wrap(self.__da << self.__unwrap(other))

    def __rlshift__(self, other):
        # Not supported by xarray, will raise
        return self.__wrap(self.__unwrap(other) << self.__da)

    def __rshift__(self, other):
        return self.__wrap(self.__da >> self.__unwrap(other))

    def __rrshift__(self, other):
        # Not supported by xarray, will raise
        return self.__wrap(self.__unwrap(other) >> self.__da)

    def __and__(self, other):
        return self.__wrap(self.__da & self.__unwrap(other))

    def __rand__(self, other):
        return self.__wrap(self.__unwrap(other) & self.__da)

    def __xor__(self, other):
        return self.__wrap(self.__da ^ self.__unwrap(other))

    def __rxor__(self, other):
        return self.__wrap(self.__unwrap(other) ^ self.__da)

    def __or__(self, other):
        return self.__wrap(self.__da | self.__unwrap(other))

    def __ror__(self, other):
        return self.__wrap(self.__unwrap(other) | self.__da)

    ################################################
    # Unary operations

    def __pos__(self):
        return self.__wrap(+self.__da)

    def __neg__(self):
        return self.__wrap(-self.__da)

    def __invert__(self):
        return self.__wrap(~self.__da)

    def __abs__(self):
        return self.__wrap(abs(self.__da))

    ################################################
    # Internal helpers

    @staticmethod
    def __unwrap(v):
        return v.__da if isinstance(v, ExprVar) else v

    @staticmethod
    def __wrap(v):
        return ExprVar(v) if isinstance(v, xr.DataArray) else v

    @staticmethod
    def _wrap_fn(fn: Callable) -> Callable:
        def wrapped_fn(*args, **kwargs):
            return ExprVar.__wrap(
                fn(
                    *(ExprVar.__unwrap(arg) for arg in args),
                    **{kw: ExprVar.__unwrap(arg) for kw, arg in kwargs.items()},
                )
            )

        return wrapped_fn



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
import inspect
import itertools
import json
import math
import threading
import warnings
from collections.abc import Iterator, Sequence
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numcodecs.abc
import numpy as np
import xarray as xr
import zarr.storage

from xcube.util.assertions import assert_instance, assert_true

GetData = Callable[[tuple[int]], Union[bytes, np.ndarray]]

OnClose = Callable[[dict[str, Any]], None]


class GenericArray(dict[str, any]):
    """Represent a generic array in the ``GenericZarrStore`` as
    dictionary of properties.

    Although all properties of this class are optional,
    some of them are mandatory when added to the ``GenericZarrStore``.

    When added to the store using ``GenericZarrStore.add_array()``,
    the array *name* and *dims* must always be present.
    Other mandatory properties depend on
    the *data* and *get_data* properties, which are mutually exclusive:

    * *get_data* is called for a requested data chunk of an array.
      It must return a bytes object or a numpy nd-array and is passed
      the chunk index, the chunk shape, and this array info dictionary.
      *get_data* requires the following properties to be present too:
      *name*, *dims*, *dtype*, *shape*.
      *chunks* is optional and defaults to *shape*.
    * *data* must be a bytes object or a numpy nd-array.
      *data* requires the following properties to be present too:
      *name*, *dims*. *chunks* must be same as *shape*.

    The function *get_data* receives only keyword-arguments which
    comprises the ones passed by *get_data_params*, if any, and
    two special ones which may occur in the signature of *get_data*:

    * The keyword argument *chunk_info*, if given, provides a dictionary
      that holds information about the current chunk:
      - ``index: tuple[int, ...]`` - the chunk's index
      - ``shape: tuple[int, ...]`` - the chunk's shape
      - ``slices: tuple[slice, ...]`` - the chunk's array slices

    * The keyword argument *array_info*, if given, provides a dictionary
      that holds information about the overall array. It contains
      all array properties passed to the constructor of ``GenericArray``
      plus
      - ``ndim: int`` - number of dimensions
      - ``num_chunks: tuple[int, ...]`` - number of chunks in every dimension

    ``GenericZarrStore`` will convert a Numpy array returned
    by *get_data* or given by *data* into a bytes object.
    It will also be compressed, if a *compressor* is given.
    It is important that the array chunks always See also
    https://zarr.readthedocs.io/en/stable/spec/v2.html#chunks

    Note that if the value of a named keyword argument is None,
    it will not be stored.

    Args:
        array: Optional array info dictionary
        name: Optional array name
        data: Optional array data. Mutually exclusive with *get_data*.
            Must be a bytes object or a numpy array.
        get_data: Optional array data chunk getter. Mutually exclusive
            with *data*. Called for a requested data chunk of an array.
            Must return a bytes object or a numpy array.
        get_data_params: Optional keyword-arguments passed to
            *get_data*.
        dtype: Optional array data type. Either a string using syntax of
            the Zarr spec or a ``numpy.dtype``. For string encoded data
            types, see
            https://zarr.readthedocs.io/en/stable/spec/v2.html#data-
            type-encoding
        dims: Optional sequence of dimension names.
        shape: Optional sequence of shape sizes for each dimension.
        chunks: Optional sequence of chunk sizes for each dimension.
        fill_value: Optional fill value, see
            https://zarr.readthedocs.io/en/stable/spec/v2.html#fill-
            value-encoding
        compressor: Optional compressor. If given, it must be an
            instance of ``numcodecs.abc.Codec``.
        filters: Optional sequence of filters, see
            https://zarr.readthedocs.io/en/stable/spec/v2.html#filters.
        order: Optional array endian ordering. If given, must be "C" or
            "F". Defaults to "C".
        attrs: Optional array attributes. If given, must be JSON-
            serializable.
        on_close: Optional array close handler. Called if the store is
            closed.
        chunk_encoding: Optional encoding type of the chunk data
            returned for the array. Can be "bytes" (the default) or
            "ndarray" for array chunks that are numpy.ndarray instances.
        kwargs: Other keyword arguments passed directly to the
            dictionary constructor.
    """

    def __init__(
        self,
        array: Optional[dict[str, any]] = None,
        name: Optional[str] = None,
        get_data: Optional[GetData] = None,
        get_data_params: Optional[dict[str, Any]] = None,
        data: Optional[np.ndarray] = None,
        dtype: Optional[Union[str, np.dtype]] = None,
        dims: Optional[Union[str, Sequence[str]]] = None,
        shape: Optional[Sequence[int]] = None,
        chunks: Optional[Sequence[int]] = None,
        fill_value: Optional[Union[bool, int, float, str]] = None,
        compressor: Optional[numcodecs.abc.Codec] = None,
        filters: Optional[Sequence[numcodecs.abc.Codec]] = None,
        order: Optional[str] = None,
        attrs: Optional[dict[str, Any]] = None,
        on_close: Optional[OnClose] = None,
        chunk_encoding: Optional[str] = None,
        **kwargs,
    ):
        array = dict(array) if array is not None else dict()
        array.update(
            {
                k: v
                for k, v in dict(
                    name=name,
                    dtype=dtype,
                    dims=dims,
                    shape=shape,
                    chunks=chunks,
                    fill_value=fill_value,
                    compressor=compressor,
                    filters=filters,
                    order=order,
                    attrs=attrs,
                    data=data,
                    get_data=get_data,
                    get_data_params=get_data_params,
                    on_close=on_close,
                    chunk_encoding=chunk_encoding,
                ).items()
                if v is not None
            }
        )
        super().__init__(array, **kwargs)

    def finalize(self) -> "GenericArray":
        """Normalize and validate array properties and return a valid
        array info dictionary to be stored in the `GenericZarrStore`.
        """
        name = self.get("name")
        if not name:
            raise ValueError("missing array name")

        data = self.get("data")
        get_data = self.get("get_data")
        if data is None and get_data is None:
            raise ValueError(f"array {name!r}: either data or get_data must be defined")
        if get_data is not None:
            if data is not None:
                raise ValueError(
                    f"array {name!r}: data and get_data cannot be defined together"
                )
            if not callable(get_data):
                raise TypeError(f"array {name!r}: get_data must be a callable")
            sig = inspect.signature(get_data)
            get_data_info = {
                "has_array_info": "array_info" in sig.parameters,
                "has_chunk_info": "chunk_info" in sig.parameters,
            }
            get_data_params = dict(self.get("get_data_params") or {})
        else:
            get_data_info = None
            get_data_params = None

        dims = self.get("dims")
        dims = [dims] if isinstance(dims, str) else dims
        if dims is None:
            raise ValueError(f"array {name!r}: missing dims")

        ndim = len(dims)

        if isinstance(data, np.ndarray):
            # forman: maybe warn if dtype or shape is given,
            #  but does not match data.dtype and data.shape
            dtype = str(data.dtype.str)
            shape = data.shape
            chunks = data.shape
        else:
            dtype = self.get("dtype")
            shape = self.get("shape")
            chunks = self.get("chunks", shape)

        if not dtype:
            raise ValueError(f"array {name!r}: missing dtype")
        elif isinstance(dtype, np.dtype):
            dtype = dtype.str

        if shape is None:
            raise ValueError(f"array {name!r}: missing shape")
        if len(shape) != ndim:
            raise ValueError(f"array {name!r}: dims and shape must have same length")
        if len(chunks) != ndim:
            raise ValueError(f"array {name!r}: dims and chunks must have same length")

        num_chunks = tuple(map(lambda x: math.ceil(x[0] / x[1]), zip(shape, chunks)))

        filters = self.get("filters")
        if filters:
            filters = list(filters)
            for f in filters:
                if not isinstance(f, numcodecs.abc.Codec):
                    raise TypeError(
                        f"array {name!r}:"
                        f" filter items must be an"
                        f" instance of numcodecs.abc.Codec"
                    )
        else:
            filters = None

        compressor = self.get("compressor")
        if compressor is not None:
            if not isinstance(compressor, numcodecs.abc.Codec):
                raise TypeError(
                    f"array {name!r}:"
                    f" compressor must be an"
                    f" instance of numcodecs.abc.Codec"
                )

        fill_value = self.get("fill_value")
        if isinstance(fill_value, np.ndarray):
            fill_value = fill_value.item()
        allowed_fill_value_types = (type(None), bool, int, float, str)
        if not isinstance(fill_value, allowed_fill_value_types):
            raise TypeError(
                f"array {name!r}:"
                f" fill_value type must be one of"
                f" {tuple(t.__name__ for t in allowed_fill_value_types)},"
                f" was {type(fill_value).__name__}"
            )

        order = self.get("order") or "C"
        allowed_orders = ("C", "F")
        if order not in allowed_orders:
            raise ValueError(
                f"array {name!r}: order must be one of {allowed_orders}, was {order!r}"
            )

        chunk_encoding = self.get("chunk_encoding") or "bytes"
        allowed_chunk_encodings = ("bytes", "ndarray")
        if chunk_encoding not in allowed_chunk_encodings:
            raise ValueError(
                f"array {name!r}:"
                f" chunk_encoding must be one of {allowed_chunk_encodings},"
                f" was {chunk_encoding!r}"
            )

        attrs = self.get("attrs")
        if attrs is not None:
            if not isinstance(attrs, dict):
                raise TypeError(
                    f"array {name!r}: attrs must be dict, was {type(attrs).__name__}"
                )

        # Note: passing the properties as dictionary
        # will prevent removing them if their value is None,
        # see GenericArray constructor.
        return GenericArray(
            {
                "name": name,
                "dtype": dtype,
                "dims": tuple(dims),
                "shape": tuple(shape),
                "chunks": tuple(chunks),
                "fill_value": fill_value,
                "filters": filters,
                "compressor": compressor,
                "order": order,
                "attrs": attrs,
                "data": data,
                "get_data": get_data,
                "get_data_params": get_data_params,
                "on_close": self.get("on_close"),
                "chunk_encoding": chunk_encoding,
                # Computed properties
                "ndim": len(dims),
                "num_chunks": num_chunks,
                "get_data_info": get_data_info,
            }
        )


GenericArrayLike = Union[GenericArray, dict[str, Any]]


class GenericZarrStore(zarr.storage.Store):
    """A Zarr store that maintains generic arrays in a flat, top-level
    hierarchy. The root of the store is a Zarr group
    conforming to the Zarr spec v2.

    It is designed to serve as a Zarr store for xarray datasets
    that compute their data arrays dynamically.

    See class ``GenericArray`` for specifying the arrays' properties.

    The array data of this store's arrays are either retrieved from
    static (numpy) arrays or from a callable that provides the
    array's data chunks as bytes or numpy arrays.

    Args:
        arrays: Arrays to be added.
            Typically, these will be instances of ``GenericArray``.
        attrs: Optional attributes of the top-level group.
            If given, it must be JSON serializable.
        array_defaults: Optional array defaults for
            array properties not passed to ``add_array``.
            Typically, this will be an instance of ``GenericArray``.
    """

    # Shortcut for GenericArray
    Array = GenericArray

    def __init__(
        self,
        *arrays: GenericArrayLike,
        attrs: Optional[dict[str, Any]] = None,
        array_defaults: Optional[GenericArrayLike] = None,
    ):
        self._attrs = dict(attrs) if attrs is not None else {}
        self._array_defaults = array_defaults
        self._dim_sizes: dict[str, int] = {}
        self._arrays: dict[str, GenericArray] = {}
        for array in arrays:
            self.add_array(array)

    def add_array(
        self, array: Optional[GenericArrayLike] = None, **array_kwargs
    ) -> None:
        """Add a new array to this store.

        Args:
            array: Optional array properties.
                Typically, this will be an instance of ``GenericArray``.
            array_kwargs: Keyword arguments form
                for the properties of ``GenericArray``.
        """
        effective_array = GenericArray(self._array_defaults or {})
        if array:
            effective_array.update(array)
        if array_kwargs:
            effective_array.update(array_kwargs)
        effective_array = effective_array.finalize()

        name = effective_array["name"]
        if name in self._arrays:
            raise ValueError(f"array {name!r} is already defined")

        dims = effective_array["dims"]
        shape = effective_array["shape"]
        for dim_name, dim_size in zip(dims, shape):
            old_dim_size = self._dim_sizes.get(dim_name)
            if old_dim_size is None:
                self._dim_sizes[name] = dim_size
            elif old_dim_size != dim_size:
                # Dimensions must have same lengths for all arrays
                # in this store
                raise ValueError(
                    f"array {name!r}"
                    f" defines dimension {dim_name!r}"
                    f" with size {dim_size},"
                    f" but existing size is {old_dim_size}"
                )

        self._arrays[name] = effective_array

    ##########################################################################
    # Zarr Store implementation
    ##########################################################################

    def is_writeable(self) -> bool:
        """Return False, because arrays in this store are generative."""
        return False

    def listdir(self, path: str = "") -> list[str]:
        """List a store path.

        Args:
            path: The path.

        Returns: List of sorted directory entries.
        """
        if path == "":
            return sorted([".zmetadata", ".zgroup", ".zattrs", *self._arrays.keys()])
        elif "/" not in path:
            return sorted(self._get_array_keys(path))
        raise ValueError(f"{path} is not a directory")

    def rmdir(self, path: str = "") -> None:
        """The general form removes store paths.
        This implementation can remove entire arrays only.

        Args:
            path: The array's name.
        """
        if path not in self._arrays:
            raise ValueError(f"{path}: can only remove existing arrays")
        array = self._arrays.pop(path)
        dims = array["dims"]
        for i, dim_name in enumerate(dims):
            dim_used = False
            for array_name, array in self._arrays.items():
                if dim_name in array["dims"]:
                    dim_used = True
                    break
            if not dim_used:
                del self._dim_sizes[dim_name]

    def rename(self, src_path: str, dst_path: str) -> None:
        """The general form renames store paths.
        This implementation can rename arrays only.

        Args:
            src_path: Source array name.
            dst_path: Target array name.
        """
        array = self._arrays.get(src_path)
        if array is None:
            raise ValueError(
                f"can only rename arrays, but {src_path!r} is not an array"
            )
        if dst_path in self._arrays:
            raise ValueError(
                f"cannot rename array {src_path!r} into"
                f" {dst_path!r} because it already exists"
            )
        if "/" in dst_path:
            raise ValueError(f"cannot rename array {src_path!r} into {dst_path!r}")
        array["name"] = dst_path
        self._arrays[dst_path] = array
        del self._arrays[src_path]

    def close(self) -> None:
        """Calls the "on_close" handlers, if any, of arrays."""
        for array in self._arrays.values():
            on_close = array.get("on_close")
            if on_close is not None:
                on_close(array)

    # Note, getsize is not implemented by intention as it requires
    # actual computation of arrays.
    #
    # def getsize(self, key: str) -> int:
    #    pass

    ##########################################################################
    # MutableMapping implementation
    ##########################################################################

    def __iter__(self) -> Iterator[str]:
        """Get an iterator of all keys in this store."""
        yield ".zmetadata"
        yield ".zgroup"
        yield ".zattrs"
        for array_name in self._arrays.keys():
            yield from self._get_array_keys(array_name)

    def __len__(self) -> int:
        return sum(1 for _ in iter(self))

    def __contains__(self, key: str) -> bool:
        if key in (".zmetadata", ".zgroup", ".zattrs"):
            return True
        try:
            array_name, value_id = self._parse_array_key(key)
        except KeyError:
            return False
        if value_id in (".zarray", ".zattrs"):
            return True
        try:
            self._get_array_chunk_index(array_name, value_id)
            return True
        except KeyError:
            return False

    def __getitem__(self, key: str) -> Union[bytes, np.ndarray]:
        item = self._get_item(key)
        if isinstance(item, dict):
            return dict_to_bytes(item)
        elif isinstance(item, str):
            return str_to_bytes(item)
        return item

    def __setitem__(self, key: str, value: bytes) -> None:
        class_name = self.__module__ + "." + self.__class__.__name__
        raise TypeError(f"{class_name} is read-only")

    def __delitem__(self, key: str) -> None:
        self.rmdir(key)

    ########################################################################
    # Utilities
    ##########################################################################

    @classmethod
    def from_dataset(
        cls, dataset: xr.Dataset, array_defaults: Optional[GenericArrayLike] = None
    ) -> "GenericZarrStore":
        """Create a Zarr store for given *dataset*.
        to the *dataset*'s attributes.
        The following *array_defaults* properties can be provided
        (other properties are prescribed by the *dataset*):

        * ``fill_value``- defaults to None
        * ``compressor``- defaults to None
        * ``filters``- defaults to None
        * ``order``- defaults to "C"
        * ``chunk_encoding`` - defaults to "bytes"

        Args:
            dataset: The dataset
            array_defaults: Array default values.

        Returns: A new Zarr store instance.
        """

        def _get_dataset_data(ds=None, chunk_info=None, array_info=None) -> np.ndarray:
            array_name = array_info["name"]
            chunk_slices = chunk_info["slices"]
            return ds[array_name][chunk_slices].values

        arrays = []
        for var_name, var in dataset.variables.items():
            arrays.append(
                GenericArray(
                    name=str(var_name),
                    dtype=np.dtype(var.dtype).str,
                    dims=[str(dim) for dim in var.dims],
                    shape=var.shape,
                    chunks=(
                        [(max(*c) if len(c) > 1 else c[0]) for c in var.chunks]
                        if var.chunks
                        else var.shape
                    ),
                    attrs={str(k): v for k, v in var.attrs.items()},
                    get_data=_get_dataset_data,
                    get_data_params=dict(ds=dataset),
                )
            )

        attrs = {str(k): v for k, v in dataset.attrs.items()}
        return GenericZarrStore(*arrays, attrs=attrs, array_defaults=array_defaults)

    ########################################################################
    # Helpers
    ##########################################################################

    def _get_item(self, key: str) -> Union[dict, str, bytes]:
        if key == ".zmetadata":
            return self._get_metadata_item()
        if key == ".zgroup":
            return self._get_group_item()
        if key == ".zattrs":
            return self._get_attrs_item()

        array_name, value_id = self._parse_array_key(key)
        array = self._arrays[array_name]

        if value_id == ".zarray":
            return self._get_array_spec_item(array)
        if value_id == ".zattrs":
            return self._get_array_attrs_item(array)

        chunk_index = self._get_array_chunk_index(array_name, value_id)
        return self._get_array_data_item(array, chunk_index)

    def _get_metadata_item(self):
        metadata = {
            ".zgroup": self._get_item(".zgroup"),
            ".zattrs": self._get_item(".zattrs"),
        }
        for array_name in self._arrays.keys():
            key = array_name + "/.zarray"
            metadata[key] = self._get_item(key)
            key = array_name + "/.zattrs"
            metadata[key] = self._get_item(key)
        return {"zarr_consolidated_format": 1, "metadata": metadata}

    # noinspection PyMethodMayBeStatic
    def _get_group_item(self):
        return {"zarr_format": 2}

    def _get_attrs_item(self):
        return self._attrs or {}

    # noinspection PyMethodMayBeStatic
    def _get_array_spec_item(self, array: GenericArray):
        # JSON-encode fill_value
        fill_value = array["fill_value"]
        if isinstance(fill_value, float):
            if math.isnan(fill_value):
                fill_value = "NaN"
            elif math.isinf(fill_value):
                if fill_value < 0:
                    fill_value = "-Infinity"
                else:
                    fill_value = "Infinity"

        # JSON-encode compressor
        compressor = array["compressor"]
        if compressor is not None:
            compressor = compressor.get_config()

        # JSON-encode filters
        filters = array["filters"]
        if filters is not None:
            filters = list(f.get_config() for f in filters)

        return {
            "zarr_format": 2,
            "dtype": array["dtype"],
            "shape": list(array["shape"]),
            "chunks": list(array["chunks"]),
            "fill_value": fill_value,
            "compressor": compressor,
            "filters": filters,
            "order": array["order"],
        }

    # noinspection PyMethodMayBeStatic
    def _get_array_attrs_item(self, array: GenericArray):
        dims = array["dims"]
        attrs = array["attrs"]
        return {"_ARRAY_DIMENSIONS": dims, **(attrs or {})}

    # noinspection PyMethodMayBeStatic
    def _get_array_data_item(
        self, array: dict[str, Any], chunk_index: tuple[int]
    ) -> Union[bytes, np.ndarray]:
        # Note, here array is expected to be "finalized",
        # that is, validated and normalized

        shape = array["shape"]
        chunks = array["chunks"]
        chunk_shape = None

        data = array["data"]
        if data is None:
            get_data = array["get_data"]
            assert callable(get_data)  # Has been ensured before
            get_data_params = array["get_data_params"]
            get_data_kwargs = dict(get_data_params)
            get_data_info = array["get_data_info"]
            if get_data_info["has_chunk_info"]:
                chunk_shape = get_chunk_shape(shape, chunks, chunk_index)
                array_slices = get_array_slices(shape, chunks, chunk_index)
                get_data_kwargs["chunk_info"] = {
                    "index": chunk_index,
                    "shape": chunk_shape,
                    "slices": array_slices,
                }
            if get_data_info["has_array_info"]:
                get_data_kwargs["array_info"] = dict(array)

            data = get_data(**get_data_kwargs)

        chunk_encoding = array["chunk_encoding"]
        if isinstance(data, np.ndarray):
            # As of Zarr 2.0, all chunks of an array
            # must have the same shape (= chunks)
            if data.shape != chunks:
                # This commonly happens if array shape sizes
                # are not integer multiple of chunk shape sizes.
                if chunk_shape is None:
                    # Compute expected chunk shape.
                    chunk_shape = get_chunk_shape(shape, chunks, chunk_index)
                # We will only pad the data if the data shape
                # corresponds to the expected chunk's shape.
                if data.shape == chunk_shape:
                    padding = get_chunk_padding(shape, chunks, chunk_index)
                    fill_value = array["fill_value"]
                    data = np.pad(
                        data, padding, mode="constant", constant_values=fill_value or 0
                    )
                else:
                    key = format_chunk_key(array["name"], chunk_index)
                    raise ValueError(
                        f"{key}:"
                        f" data chunk at {chunk_index}"
                        f" must have shape {chunk_shape},"
                        f" but was {data.shape}"
                    )
            if chunk_encoding == "bytes":
                # Convert to bytes, filter and compress
                data = ndarray_to_bytes(
                    data,
                    order=array["order"],
                    filters=array["filters"],
                    compressor=array["compressor"],
                )

        # Sanity check
        if (chunk_encoding == "bytes" and not isinstance(data, bytes)) or (
            chunk_encoding == "ndarray" and not isinstance(data, np.ndarray)
        ):
            key = format_chunk_key(array["name"], chunk_index)
            expected_type = "numpy.ndarray" if chunk_encoding == "ndarray" else "bytes"
            raise TypeError(
                f"{key}:"
                f" data must be encoded as {expected_type},"
                f" but was {type(data).__name__}"
            )

        return data

    def _parse_array_key(self, key: str) -> tuple[str, str]:
        array_name_and_value_id = key.rsplit("/", maxsplit=1)
        if len(array_name_and_value_id) != 2:
            raise KeyError(key)
        array_name, value_id = array_name_and_value_id
        if array_name not in self._arrays:
            raise KeyError(key)
        return array_name, value_id

    def _get_array_chunk_index(self, array_name: str, index_id: str) -> tuple[int]:
        try:
            chunk_index = tuple(map(int, index_id.split(".")))
        except (ValueError, TypeError):
            raise KeyError(f"{array_name}/{index_id}")
        array = self._arrays[array_name]
        shape = array["shape"]
        if len(chunk_index) != len(shape):
            raise KeyError(f"{array_name}/{index_id}")
        num_chunks = array["num_chunks"]
        for i, n in zip(chunk_index, num_chunks):
            if not (0 <= i < n):
                raise KeyError(f"{array_name}/{index_id}")
        return chunk_index

    def _get_array_keys(self, array_name: str) -> Iterator[str]:
        yield array_name + "/.zarray"
        yield array_name + "/.zattrs"
        array = self._arrays[array_name]
        num_chunks = array["num_chunks"]
        yield from get_chunk_keys(array_name, num_chunks)


def get_array_slices(
    shape: tuple[int, ...], chunks: tuple[int, ...], chunk_index: tuple[int, ...]
) -> tuple[slice, ...]:
    return tuple(
        slice(i * c, i * c + (c if (i + 1) * c <= s else s % c))
        for s, c, i in zip(shape, chunks, chunk_index)
    )


def get_chunk_shape(
    shape: tuple[int, ...], chunks: tuple[int, ...], chunk_index: tuple[int, ...]
) -> tuple[int, ...]:
    return tuple(
        c if (i + 1) * c <= s else s % c for s, c, i in zip(shape, chunks, chunk_index)
    )


def get_chunk_padding(
    shape: tuple[int, ...], chunks: tuple[int, ...], chunk_index: tuple[int, ...]
):
    return tuple(
        (0, 0 if (i + 1) * c <= s else c - s % c)
        for s, c, i in zip(shape, chunks, chunk_index)
    )


def get_chunk_indexes(num_chunks: tuple[int, ...]) -> Iterator[tuple[int, ...]]:
    if not num_chunks:
        yield (0,)
    else:
        yield from itertools.product(*tuple(map(range, map(int, num_chunks))))


def get_chunk_keys(array_name: str, num_chunks: tuple[int, ...]) -> Iterator[str]:
    for chunk_index in get_chunk_indexes(num_chunks):
        yield format_chunk_key(array_name, chunk_index)


def format_chunk_key(array_name: str, chunk_index: tuple[int, ...]) -> str:
    chunk_id = ".".join(map(str, chunk_index))
    return f"{array_name}/{chunk_id}"


def dict_to_bytes(d: dict) -> bytes:
    return str_to_bytes(json.dumps(d, indent=2))


def str_to_bytes(s: str) -> bytes:
    return bytes(s, encoding="utf-8")


def ndarray_to_bytes(
    data: np.ndarray,
    order: Optional[str] = None,
    filters: Optional[Sequence[Any]] = None,
    compressor: Optional[numcodecs.abc.Codec] = None,
) -> bytes:
    data = data.tobytes(order=order or "C")
    if filters:
        for f in filters:
            data = f.encode(data)
    if compressor is not None:
        data = compressor.encode(data)
    return data



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import inspect
import warnings
from collections.abc import Sequence
from typing import AbstractSet, Any, Callable, Dict, Tuple, Union

import numpy as np
import xarray as xr

from xcube.core.chunkstore import ChunkStore
from xcube.core.schema import CubeSchema
from xcube.core.verify import assert_cube

CubeFuncOutput = Union[
    xr.DataArray, np.ndarray, Sequence[Union[xr.DataArray, np.ndarray]]
]
CubeFunc = Callable[..., CubeFuncOutput]

_PREDEFINED_KEYWORDS = ["input_params", "dim_coords", "dim_ranges"]


# TODO: support vectorize = all cubes have same variables and cube_func receives variables as vectors (with extra dim)


def compute_cube(
    cube_func: CubeFunc,
    *input_cubes: xr.Dataset,
    input_cube_schema: CubeSchema = None,
    input_var_names: Sequence[str] = None,
    input_params: dict[str, Any] = None,
    output_var_name: str = "output",
    output_var_dtype: Any = np.float64,
    output_var_attrs: dict[str, Any] = None,
    vectorize: bool = None,
    cube_asserted: bool = False,
) -> xr.Dataset:
    """Compute a new output data cube with a single variable named
    *output_var_name* from variables named *input_var_names* contained in
    zero, one, or more input data cubes in *input_cubes* using a cube
    factory function *cube_func*.

    For a more detailed description of the function usage,
    please refer to :func:`compute_dataset`.

    Args:
        cube_func: The cube factory function.
        input_cubes: An optional sequence of input cube datasets,
            must be provided if *input_cube_schema* is not.
        input_cube_schema: An optional input cube schema,
            must be provided if *input_cubes* is not.
            Will be ignored if *input_cubes* is provided.
        input_var_names: A sequence of variable names
        input_params: Optional dictionary with processing parameters
            passed to *cube_func*.
        output_var_name: Optional name of the output variable,
            defaults to ``'output'``.
        output_var_dtype: Optional numpy datatype of the output variable,
            defaults to ``'float32'``.
        output_var_attrs: Optional metadata attributes for the output variable.
        vectorize: Whether all *input_cubes* have the same variables which
            are concatenated and passed as vectors
            to *cube_func*. Not implemented yet.
        cube_asserted: If False, *cube* will be verified,
            otherwise it is expected to be a valid cube.
    Returns:
        A new dataset that contains the computed output variable.
    """
    return compute_dataset(
        cube_func,
        *input_cubes,
        input_cube_schema=input_cube_schema,
        input_var_names=input_var_names,
        input_params=input_params,
        output_var_name=output_var_name,
        output_var_dtype=output_var_dtype,
        output_var_attrs=output_var_attrs,
        vectorize=vectorize,
        cube_asserted=cube_asserted,
    )


def compute_dataset(
    cube_func: CubeFunc,
    *input_cubes: xr.Dataset,
    input_cube_schema: CubeSchema = None,
    input_var_names: Sequence[str] = None,
    input_params: dict[str, Any] = None,
    output_var_name: str = "output",
    output_var_dims: AbstractSet[str] = None,
    output_var_dtype: Any = np.float64,
    output_var_attrs: dict[str, Any] = None,
    vectorize: bool = None,
    cube_asserted: bool = False,
) -> xr.Dataset:
    """Compute a new output dataset with a single variable named *output_var_name*
    from variables named *input_var_names* contained in zero, one, or more
    input data cubes in *input_cubes* using a cube factory function *cube_func*.

    *cube_func* is called concurrently for each of the chunks of the input variables.
    It is expected to return a chunk block whith is type ``np.ndarray``.

    If *input_cubes* is not empty, *cube_func* receives variables as specified by *input_var_names*.
    If *input_cubes* is empty, *input_var_names* must be empty too, and *input_cube_schema*
    must be given, so that a new cube can be created.

    The full signature of *cube_func* is:::

        def cube_func(*input_vars: np.ndarray,
                      input_params: Dict[str, Any] = None,
                      dim_coords: Dict[str, np.ndarray] = None,
                      dim_ranges: Dict[str, Tuple[int, int]] = None) -> np.ndarray:
            pass

    The arguments are:

    * ``input_vars``: the variables according to the given *input_var_names*;
    * ``input_params``: is this call's *input_params*, a mapping from parameter name to value;
    * ``dim_coords``: a mapping from dimension names to the current chunk's coordinate arrays;
    * ``dim_ranges``: a mapping from dimension names to the current chunk's index ranges.

    Only the ``input_vars`` argument is mandatory. The keyword arguments
    ``input_params``, ``input_params``, ``input_params`` do need to be present at all.

    *output_var_dims* my be given in the case, where ...
    TODO: describe new output_var_dims...

    Args:
        cube_func: The cube factory function.
        *input_cubes: An optional sequence of input cube datasets, must
            be provided if *input_cube_schema* is not.
        input_cube_schema: An optional input cube schema, must be
            provided if *input_cubes* is not.
        input_var_names: A sequence of variable names
        input_params: Optional dictionary with processing parameters
            passed to *cube_func*.
        output_var_name: Optional name of the output variable, defaults
            to ``'output'``.
        output_var_dims: Optional set of names of the output dimensions,
            used in the case *cube_func* reduces dimensions.
        output_var_dtype: Optional numpy datatype of the output
            variable, defaults to ``'float32'``.
        output_var_attrs: Optional metadata attributes for the output
            variable.
        vectorize: Whether all *input_cubes* have the same variables
            which are concatenated and passed as vectors to *cube_func*.
            Not implemented yet.
        cube_asserted: If False, *cube* will be verified, otherwise it
            is expected to be a valid cube.

    Returns:
        A new dataset that contains the computed output variable.
    """
    if vectorize is not None:
        # TODO: support vectorize = all cubes have same variables and cube_func
        #       receives variables as vectors (with extra dim)
        raise NotImplementedError("vectorize is not supported yet")

    if not cube_asserted:
        for cube in input_cubes:
            assert_cube(cube)

    # Check compatibility of inputs
    if input_cubes:
        input_cube_schema = CubeSchema.new(input_cubes[0])
        for cube in input_cubes:
            if not cube_asserted:
                assert_cube(cube)
            if cube != input_cubes[0]:
                # noinspection PyUnusedLocal
                other_schema = CubeSchema.new(cube)
                # TODO (forman): broadcast all cubes to same shape, rechunk to same chunks
    elif input_cube_schema is None:
        raise ValueError("input_cube_schema must be given")

    output_var_name = output_var_name or "output"

    # Collect named input variables, raise if not found
    input_var_names = input_var_names or []
    input_vars = []
    for var_name in input_var_names:
        input_var = None
        for cube in input_cubes:
            if var_name in cube.data_vars:
                input_var = cube[var_name]
                break
        if input_var is None:
            raise ValueError(f"variable {var_name!r} not found in any of cubes")
        input_vars.append(input_var)

    # Find out, if cube_func uses any of _PREDEFINED_KEYWORDS
    has_input_params, has_dim_coords, has_dim_ranges = _inspect_cube_func(
        cube_func, input_var_names
    )

    def cube_func_wrapper(index_chunk, *input_var_chunks):
        nonlocal input_cube_schema, input_var_names, input_params, input_vars
        nonlocal has_input_params, has_dim_coords, has_dim_ranges

        # Note, xarray.apply_ufunc does a test call with empty input arrays,
        # so index_chunk.size == 0 is a valid case
        empty_call = index_chunk.size == 0

        # TODO: when output_var_dims is given, index_chunk must be reordered
        #   as core dimensions are moved to the and of index_chunk and input_var_chunks
        if not empty_call:
            index_chunk = index_chunk.ravel()

        if index_chunk.size < 2 * input_cube_schema.ndim:
            if not empty_call:
                warnings.warn(
                    f"unexpected index_chunk of size {index_chunk.size} received!"
                )
                return None

        dim_ranges = None
        if has_dim_ranges or has_dim_coords:
            dim_ranges = {}
            for i in range(input_cube_schema.ndim):
                dim_name = input_cube_schema.dims[i]
                if not empty_call:
                    start = int(index_chunk[2 * i + 0])
                    end = int(index_chunk[2 * i + 1])
                    dim_ranges[dim_name] = start, end
                else:
                    dim_ranges[dim_name] = ()

        dim_coords = None
        if has_dim_coords:
            dim_coords = {}
            for coord_var_name, coord_var in input_cube_schema.coords.items():
                coord_slices = [slice(None)] * coord_var.ndim
                for i in range(input_cube_schema.ndim):
                    dim_name = input_cube_schema.dims[i]
                    if dim_name in coord_var.dims:
                        j = coord_var.dims.index(dim_name)
                        coord_slices[j] = slice(*dim_ranges[dim_name])
                dim_coords[coord_var_name] = coord_var[tuple(coord_slices)].values

        kwargs = {}
        if has_input_params:
            kwargs["input_params"] = input_params
        if has_dim_ranges:
            kwargs["dim_ranges"] = dim_ranges
        if has_dim_coords:
            kwargs["dim_coords"] = dim_coords

        return cube_func(*input_var_chunks, **kwargs)

    index_var = _gen_index_var(input_cube_schema)

    all_input_vars = [index_var] + input_vars

    input_core_dims = None
    if output_var_dims:
        input_core_dims = []
        has_warned = False
        for i in range(len(all_input_vars)):
            input_var = all_input_vars[i]
            var_core_dims = [
                dim for dim in input_var.dims if dim not in output_var_dims
            ]
            must_rechunk = False
            if var_core_dims and input_var.chunks:
                for var_core_dim in var_core_dims:
                    dim_index = input_var.dims.index(var_core_dim)
                    dim_chunk_size = input_var.chunks[dim_index][0]
                    dim_shape_size = input_var.shape[dim_index]
                    if dim_chunk_size != dim_shape_size:
                        must_rechunk = True
                        break
            if must_rechunk:
                if not has_warned:
                    warnings.warn(
                        f"Input variables must not be chunked in dimension(s): {', '.join(var_core_dims)}.\n"
                        f"Rechunking applies, which may drastically decrease runtime performance "
                        f"and increase memory usage."
                    )
                    has_warned = True
                all_input_vars[i] = input_var.chunk(
                    {var_core_dim: -1 for var_core_dim in var_core_dims}
                )
            input_core_dims.append(var_core_dims)

    output_var = xr.apply_ufunc(
        cube_func_wrapper,
        *all_input_vars,
        dask="parallelized",
        input_core_dims=input_core_dims,
        output_dtypes=[output_var_dtype],
    )
    if output_var_attrs:
        output_var.attrs.update(output_var_attrs)
    return xr.Dataset({output_var_name: output_var}, coords=input_cube_schema.coords)


def _inspect_cube_func(cube_func: CubeFunc, input_var_names: Sequence[str] = None):
    (
        args,
        varargs,
        varkw,
        defaults,
        kwonlyargs,
        kwonlydefaults,
        annotations,
    ) = inspect.getfullargspec(cube_func)
    cube_func_name = "?"
    if hasattr(cube_func, "__name__"):
        cube_func_name = cube_func.__name__
    true_args = [arg not in _PREDEFINED_KEYWORDS for arg in args]
    if False in true_args and any(true_args[true_args.index(False) :]):
        raise ValueError(
            f"invalid cube_func {cube_func_name!r}: "
            f"any argument must occur before any of {', '.join(_PREDEFINED_KEYWORDS)}, "
            f"but got {', '.join(args)}"
        )
    if not all(true_args) and varargs:
        raise ValueError(
            f"invalid cube_func {cube_func_name!r}: "
            f"any argument must occur before any of {', '.join(_PREDEFINED_KEYWORDS)}, "
            f"but got {', '.join(args)} before *{varargs}"
        )
    num_input_vars = len(input_var_names) if input_var_names else 0
    num_args = sum(true_args)
    if varargs is None and num_input_vars != num_args:
        raise ValueError(
            f"invalid cube_func {cube_func_name!r}: "
            f"expected {num_input_vars} arguments, "
            f"but got {', '.join(args)}"
        )
    has_input_params = "input_params" in args or "input_params" in kwonlyargs
    has_dim_coords = "dim_coords" in args or "dim_coords" in kwonlyargs
    has_dim_ranges = "dim_ranges" in args or "dim_ranges" in kwonlyargs
    return has_input_params, has_dim_coords, has_dim_ranges


def _gen_index_var(cube_schema: CubeSchema):
    dims = cube_schema.dims
    shape = cube_schema.shape
    chunks = cube_schema.chunks

    # noinspection PyUnusedLocal
    def get_chunk(cube_store: ChunkStore, name: str, index: tuple[int, ...]) -> bytes:
        data = np.zeros(cube_store.chunks, dtype=np.uint64)
        data_view = data.ravel()
        if data_view.base is not data:
            raise ValueError("view expected")
        if data_view.size < cube_store.ndim * 2:
            raise ValueError("size too small")
        for i in range(cube_store.ndim):
            j1 = cube_store.chunks[i] * index[i]
            j2 = j1 + cube_store.chunks[i]
            data_view[2 * i] = j1
            data_view[2 * i + 1] = j2
        return data.tobytes()

    store = ChunkStore(dims, shape, chunks)
    store.add_lazy_array("__index_var__", "<u8", get_chunk=get_chunk)

    dataset = xr.open_zarr(store)
    index_var = dataset.__index_var__
    index_var = index_var.assign_coords(**cube_schema.coords)
    return index_var



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import Api

from .context import TilesContext

api = Api(
    "tiles",
    description="xcube Tiles API",
    required_apis=["datasets"],
    create_ctx=TilesContext,
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import importlib.resources
import json
import os
import pathlib
import random
from typing import Optional, Union

from chartlets import Response as ExtResponse
from chartlets.controllers import get_callback_results, get_contributions, get_layout

from xcube.constants import LOG
from xcube.server.api import ApiError, ApiHandler
from xcube.util.jsonencoder import NumpyJSONEncoder

from .api import api
from .context import ViewerContext

ENV_VAR_XCUBE_VIEWER_PATH = "XCUBE_VIEWER_PATH"

_viewer_module = "xcube.webapi.viewer"
_data_dir = "dist"
_default_filename = "index.html"

_responses = {
    200: {
        "description": "OK",
        "content": {
            "text/html": {
                "schema": {"type": "string"},
            },
        },
    }
}


@api.static_route(
    "/viewer",
    default_filename=_default_filename,
    summary="Brings up the xcube Viewer webpage",
    responses=_responses,
)
def get_local_viewer_path() -> Union[None, str, pathlib.Path]:
    local_viewer_path = os.environ.get(ENV_VAR_XCUBE_VIEWER_PATH)
    if local_viewer_path:
        return local_viewer_path
    try:
        local_viewer_path = importlib.resources.files(_viewer_module) / _data_dir
        if (local_viewer_path / _default_filename).is_file():
            return str(local_viewer_path)
    except ImportError:
        pass
    LOG.warning(
        f"Cannot find {_data_dir}/{_default_filename}"
        f" in {_viewer_module}, consider setting environment variable"
        f" {ENV_VAR_XCUBE_VIEWER_PATH}",
        exc_info=True,
    )
    return None


@api.route("/viewer/config/{*path}")
class ViewerConfigHandler(ApiHandler[ViewerContext]):
    @api.operation(
        operationId="getViewerConfigurationItem",
        summary="Get a configuration item for the xcube viewer.",
    )
    def get(self, path: Optional[str]):
        config_items = self.ctx.config_items
        if config_items is None:
            raise ApiError.NotFound("xcube viewer has not been been configured")
        try:
            data = config_items[path]
        except KeyError:
            raise ApiError.NotFound(
                f"The item {path!r} was not found in viewer configuration"
            )
        content_type = self.get_content_type(path)
        self.response.set_header("Content-Length", str(len(data)))
        if content_type is not None:
            self.response.set_header("Content-Type", content_type)
        self.response.write(data)

    @staticmethod
    def get_content_type(path: str) -> Optional[str]:
        filename_ext = path.split("/")[-1].split(".")[-1]
        if filename_ext in ("json", "xml"):
            return f"application/{filename_ext}; charset=UTF-8"
        elif filename_ext in ("csv", "html", "css", "js"):
            return f"text/{filename_ext}; charset=UTF-8"
        elif filename_ext in ("geojson",):
            return "application/geo+json; charset=UTF-8"
        elif filename_ext in ("gif", "png"):
            return f"image/{filename_ext}"
        elif filename_ext in ("jpeg", "jpg"):
            return "image/jpeg"
        return None


@api.route("/viewer/state")
class ViewerStateHandler(ApiHandler[ViewerContext]):
    # noinspection SpellCheckingInspection
    _id_chars = "abcdefghijklmnopqrstuvwxyz0123456789"

    @api.operation(
        operationId="getViewerState",
        summary="Get previously stored viewer state keys or a specific state.",
        parameters=[
            {
                "name": "key",
                "in": "query",
                "description": "The key of a previously stored state.",
                "required": False,
                "schema": {"type": "string"},
            }
        ],
    )
    def get(self):
        if self.ctx.persistence is None:
            # 501: Not Implemented
            self.response.set_status(501, "Persistence not supported")
            return
        key = self.request.get_query_arg("key", type=str, default="")
        if key:
            if key not in self.ctx.persistence:
                self.response.set_status(404, "State not found")
                return
            state = self.ctx.persistence[key]
            LOG.info(f"Restored state ({len(state)} bytes) for key {key!r}")
            self.response.write(state)
        else:
            keys = list(self.ctx.persistence.keys())
            self.response.write({"keys": keys})

    @api.operation(
        operationId="putViewerState",
        summary="Store a viewer state and return a state key.",
    )
    def put(self):
        if self.ctx.persistence is None:
            # 501: Not Implemented
            self.response.set_status(501, "Persistence not supported")
            return
        state = self.request.body
        key = self.new_key()
        self.ctx.persistence[key] = state
        LOG.info(f"Stored state ({len(state)} bytes) using key {key!r}")
        self.response.write({"key": key})

    def new_key(self, length: int = 8) -> str:
        while True:
            key = "".join(random.choice(self._id_chars) for _ in range(length))
            if key not in self.ctx.persistence:
                return key


class ViewerExtHandler(ApiHandler[ViewerContext]):
    def do_get_contributions(self):
        if self.ctx.ext_ctx is None:
            self._write_no_ext_status()
        else:
            self._write_response(get_contributions(self.ctx.ext_ctx))

    def do_get_layout(self, contrib_point: str, contrib_index: str):
        if self.ctx.ext_ctx is None:
            self._write_no_ext_status()
        else:
            self._write_response(
                get_layout(
                    self.ctx.ext_ctx,
                    contrib_point,
                    int(contrib_index),
                    self.request.json,
                )
            )

    def do_get_callback_results(self):
        if self.ctx.ext_ctx is None:
            self._write_no_ext_status()
        else:
            self._write_response(
                get_callback_results(self.ctx.ext_ctx, self.request.json)
            )

    def _write_response(self, response: ExtResponse):
        self.response.set_header("Content-Type", "text/json")
        if response.ok:
            self.response.write(
                json.dumps({"result": response.data}, cls=NumpyJSONEncoder)
            )
        else:
            self.response.set_status(response.status, response.reason)
            self.response.write(
                {"error": {"status": response.status, "message": response.reason}}
            )

    def _write_no_ext_status(self):
        self.response.set_status(404, "No extensions configured")
        self.response.finish()


@api.route("/viewer/ext/contributions")
class ViewerExtContributionsHandler(ViewerExtHandler):
    # GET /dashi/contributions
    @api.operation(
        operationId="getViewerContributions",
        summary="Get viewer extensions and all their contributions.",
    )
    def get(self):
        self.do_get_contributions()


# noinspection PyPep8Naming
@api.route("/viewer/ext/layout/{contribPoint}/{contribIndex}")
class ViewerExtLayoutHandler(ViewerExtHandler):
    def get(self, contribPoint: str, contribIndex: str):
        """This endpoint is for testing only."""
        self.do_get_layout(contribPoint, contribIndex)

    # POST /dashi/layout/{contrib_point_name}/{contrib_index}
    @api.operation(
        operationId="getViewerContributionLayout",
        summary="Get the initial layout for the given viewer contribution.",
        parameters=[
            {
                "name": "contribPoint",
                "in": "path",
                "description": 'Contribution point name, e.g., "panels"',
                "schema": {"type": "string"},
            },
            {
                "name": "contribIndex",
                "in": "path",
                "description": "Contribution index",
                "schema": {"type": "integer"},
            },
        ],
    )
    def post(self, contribPoint: str, contribIndex: str):
        self.do_get_layout(contribPoint, contribIndex)


@api.route("/viewer/ext/callback")
class ViewerExtCallbackHandler(ViewerExtHandler):
    @api.operation(
        operationId="getViewerContributionCallbackResults",
        summary=(
            "Process the viewer contribution callback requests"
            " and return state change requests."
        ),
    )
    def post(self):
        self.do_get_callback_results()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import ApiHandler

from ..datasets import (
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    QUERY_PARAM_CMAP,
    QUERY_PARAM_CRS,
    QUERY_PARAM_NORM,
    QUERY_PARAM_VMAX,
    QUERY_PARAM_VMIN,
)
from .api import api
from .context import TilesContext
from .controllers import compute_ml_dataset_tile

PATH_PARAM_X = {
    "name": "x",
    "in": "path",
    "description": "The tile grid's x-coordinate",
    "schema": {
        "type": "integer",
    },
}

PATH_PARAM_Y = {
    "name": "y",
    "in": "path",
    "description": "The tile grid's y-coordinate",
    "schema": {
        "type": "integer",
    },
}

PATH_PARAM_Z = {
    "name": "z",
    "in": "path",
    "description": "The tile grid's z-coordinate",
    "schema": {
        "type": "integer",
    },
}

QUERY_PARAM_TIME = {
    "name": "time",
    "in": "query",
    "description": 'Optional time coordinate using format "YYYY-MM-DD hh:mm:ss"',
    "schema": {"type": "string", "format": "datetime"},
}

QUERY_PARAM_FORMAT = {
    "name": "format",
    "in": "query",
    "description": "Image format",
    "schema": {"type": "string", "enum": ["png", "image/png"], "default": "png"},
}

QUERY_PARAM_RETINA = {
    "name": "retina",
    "in": "query",
    "description": "Returns tiles of size 512 instead of 256",
    "schema": {"type": "boolean"},
}

TILE_PARAMETERS = [
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    PATH_PARAM_Z,
    PATH_PARAM_Y,
    PATH_PARAM_X,
    QUERY_PARAM_CRS,
    QUERY_PARAM_VMIN,
    QUERY_PARAM_VMAX,
    QUERY_PARAM_CMAP,
    QUERY_PARAM_NORM,
    QUERY_PARAM_TIME,
    QUERY_PARAM_FORMAT,
    QUERY_PARAM_RETINA,
]


# noinspection PyPep8Naming
@api.route("/tiles/{datasetId}/{varName}/{z}/{y}/{x}")
class TilesHandler(ApiHandler[TilesContext]):
    @api.operation(
        operation_id="getTile",
        summary="Get the image tile for a variable and given tile grid coordinates.",
        parameters=TILE_PARAMETERS,
    )
    async def get(self, datasetId: str, varName: str, z: str, y: str, x: str):
        tile = await self.ctx.run_in_executor(
            None,
            compute_ml_dataset_tile,
            self.ctx,
            datasetId,
            varName,
            None,
            x,
            y,
            z,
            {k: v[0] for k, v in self.request.query.items()},
        )
        self.response.set_header("Content-Type", "image/png")
        await self.response.finish(tile)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .context import PlacesContext
from .routes import PATH_PARAM_PLACE_GROUP_ID, api



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import abc
from collections.abc import Sequence
from typing import List, Optional, Tuple, Type

from xcube.constants import EXTENSION_POINT_SERVER_FRAMEWORKS
from xcube.util.extension import get_extension_registry
from xcube.util.jsonschema import JsonObjectSchema

from .api import ApiRoute, ApiStaticRoute, Context
from .asyncexec import AsyncExecution


class Framework(AsyncExecution, abc.ABC):
    """An abstract web server framework."""

    @property
    @abc.abstractmethod
    def config_schema(self) -> Optional[JsonObjectSchema]:
        """Returns an optional JSON Schema for
        the web server configuration. Returning None
        indicates that configuration is not possible.
        """

    @abc.abstractmethod
    def add_static_routes(self, routes: Sequence[ApiStaticRoute], url_prefix: str):
        """Adds the given static routes to this web server.

        Args:
            routes: The static routes to be added.
            url_prefix: URL prefix, may be an empty string.
        """

    @abc.abstractmethod
    def add_routes(self, routes: Sequence[ApiRoute], url_prefix: str):
        """Adds the given routes to this web server.

        Args:
            routes: The routes to be added.
            url_prefix: URL prefix, may be an empty string.
        """

    @abc.abstractmethod
    def update(self, ctx: Context):
        """Called, when the server context has changed.

        This is the case immediately after instantiation and before
        start() is called. It may then be called on any context change,
        likely due to a configuration change.

        Args:
            ctx: The current server context.
        """

    @abc.abstractmethod
    def start(self, ctx: Context):
        """Starts the web service.

        Args:
            ctx: The initial server context.
        """

    @abc.abstractmethod
    def stop(self, ctx: Context):
        """Stops the web service.

        Args:
            ctx: The current server context.
        """


def get_framework_names() -> list[str]:
    """Get the names of possible web server frameworks."""
    extension_registry = get_extension_registry()
    return [
        ext.name
        for ext in extension_registry.find_extensions(EXTENSION_POINT_SERVER_FRAMEWORKS)
    ]


def get_framework_class(framework_name: str) -> type[Framework]:
    """Get the web server framework class for the given *framework_name*."""
    extension_registry = get_extension_registry()
    return extension_registry.get_component(
        EXTENSION_POINT_SERVER_FRAMEWORKS, framework_name
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import xarray as xr

from xcube.core.gridmapping import GridMapping

from ..config import CubeConfig
from .transformer import CubeTransformer, TransformedCube


class CubeResamplerT(CubeTransformer):
    def transform_cube(
        self, cube: xr.Dataset, gm: GridMapping, cube_config: CubeConfig
    ) -> TransformedCube:
        # TODO (forman): implement me
        return cube, gm, cube_config



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import json
import sys
import traceback
from typing import Any, Dict

import click

from xcube.cli.common import (
    cli_option_quiet,
    cli_option_verbosity,
    configure_cli_output,
)
from xcube.constants import LOG


@click.command(name="gen2", hidden=True)
@click.argument("request_path", type=str, required=False, metavar="REQUEST")
@click.option(
    "--stores",
    "stores_config_path",
    metavar="STORES_CONFIG",
    help="A JSON or YAML file that maps store names to parameterized data stores.",
)
@click.option(
    "--service",
    "service_config_path",
    metavar="SERVICE_CONFIG",
    help="A JSON or YAML file that provides the configuration for an"
    " xcube Generator Service. If provided, the REQUEST will be"
    " passed to the service instead of generating the cube on "
    " this computer.",
)
@click.option(
    "--info",
    "-i",
    is_flag=True,
    help="Output information about the data cube to be generated."
    " Does not generate the data cube.",
)
@click.option(
    "--output",
    "-o",
    "output_file",
    metavar="RESULT",
    help="Write cube information or generation result into JSON"
    " file RESULT."
    " If omitted, the JSON is dumped to stdout.",
)
@cli_option_quiet
@cli_option_verbosity
def gen2(
    request_path: str,
    stores_config_path: str = None,
    service_config_path: str = None,
    output_file: str = None,
    info: bool = False,
    quiet: bool = False,
    verbosity: int = 0,
):
    """Generator tool for data cubes.

    Creates a cube view from one or more cube stores, optionally performs some
    cube transformation, and writes the resulting cube to some target cube
    store.

    REQUEST is the cube generation request. It may be provided as a JSON or
    YAML file (file extensions ".json" or ".yaml"). If the REQUEST file
    argument is omitted, it is expected that the Cube generation request is
    piped as a JSON string.

    STORE_CONFIGS is a path to a JSON or YAML file with xcube data store
    configurations. It is a mapping of arbitrary store names to configured
    data stores. Entries are dictionaries that have a mandatory "store_id"
    property which is a name of a registered xcube data store.
    The optional "store_params" property may define data store specific
    parameters. The following example defines a data store named
    "my_s3_store" which is an AWS S3 bucket store, and a data store
    "my_test_store" for testing, which is an in-memory data store:

    \b
    {
        "my_s3_store": {
            "store_id": "s3",
            "store_params": {
                "bucket_name": "eurodatacube",
                "aws_access_key_id": "jokljkjoiqqjvlaksd",
                "aws_secret_access_key": "1728349182734983248234"
            }
        },
        "my_test_store": {
            "store_id": "memory"
        }
    }

    SERVICE_CONFIG is a path to a JSON or YAML file with an xcube Generator
    service configuration. Here is a JSON example:

    \b
    {
        "endpoint_url": "https://xcube-gen.eurodatacube.com/api/v2/",
        "client_id": "ALDLPUIOSD5NHS3103",
        "client_secret": "lfaolb3klv904kj23lkdfsjkf430894341"
    }

    Values in the file may also be interpolated from current
    environment variables. In this case templates are used, for example:

    \b
    {
        "endpoint_url": "https://xcube-gen.eurodatacube.com/api/v2/",
        "client_id": "${XCUBE_GEN_CLIENT_ID}",
        "client_secret": "${XCUBE_GEN_CLIENT_SECRET}"
    }

    """
    from xcube.core.gen2 import CubeGenerator, CubeGeneratorError, CubeGeneratorRequest
    from xcube.util.versions import get_xcube_versions

    configure_cli_output(quiet=quiet, verbosity=verbosity)

    error = None

    result: dict[str, Any]

    # noinspection PyBroadException
    try:
        request = CubeGeneratorRequest.from_file(request_path, verbosity=verbosity)
        generator = CubeGenerator.new(
            stores_config=stores_config_path,
            service_config=service_config_path,
            verbosity=verbosity,
        )

        if info:
            result = generator.get_cube_info(request).to_dict()
        else:
            result = generator.generate_cube(request).to_dict()
    except BaseException as e:
        error = e

        result = dict(
            status="error",
            message=f"{error}",
            traceback=traceback.format_tb(error.__traceback__),
        )
        if isinstance(error, CubeGeneratorError):
            if error.status_code is not None:
                result.update(status_code=error.status_code)
            if error.remote_output is not None:
                result.update(remote_output=error.remote_output)
            if error.remote_traceback is not None:
                result.update(remote_traceback=error.remote_traceback)
        if "status_code" not in result:
            result["status_code"] = 500

    if result.get("status") == "error":
        result["versions"] = get_xcube_versions()

    if output_file is not None:
        with open(output_file, "w") as fp:
            json.dump(result, fp, indent=2)
        LOG.info(f"Result written to {output_file}")
    else:
        json.dump(result, sys.stdout, indent=2)

    if result.get("status") == "error":
        message = result.get(
            "message", "Cube generation failed." if error is None else f"{error}"
        )
        click_error = click.ClickException(message)
        raise click_error from error


if __name__ == "__main__":
    gen2()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
from collections.abc import Iterator

import zarr.storage


class DiagnosticZarrStore(zarr.storage.Store):
    """A diagnostic Zarr store used for testing and investigating
    behaviour of Zarr and xarray's Zarr backend.

    Args:
        store: Wrapped Zarr store.
    """

    def __init__(self, store: collections.abc.MutableMapping):
        self.store = store
        self.records: list[str] = []
        if hasattr(self.store, "listdir"):
            self.listdir = self._listdir
        if hasattr(self.store, "getsize"):
            self.getsize = self._getsize

    def _add_record(self, record: str):
        self.records.append(record)

    def _listdir(self, path: str = "") -> list[str]:
        self._add_record(f"listdir({path!r})")
        # noinspection PyUnresolvedReferences
        return self.store.listdir(path=path)

    def _getsize(self, key: str) -> int:
        self._add_record(f"getsize({key!r})")
        # noinspection PyUnresolvedReferences
        return self.store.getsize(key)

    def keys(self):
        self._add_record("keys()")
        return self.store.keys()

    def __setitem__(self, key: str, value: bytes) -> None:
        self._add_record(f"__setitem__({key!r}, {type(value).__name__})")
        self.store.__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        self._add_record(f"__delitem__({key!r})")
        self.store.__delitem__(key)

    def __getitem__(self, key: str) -> bytes:
        self._add_record(f"__getitem__({key!r})")
        return self.store.__getitem__(key)

    def __len__(self) -> int:
        self._add_record("__len__()")
        return self.store.__len__()

    def __iter__(self) -> Iterator[str]:
        self._add_record("__iter__()")
        return self.store.__iter__()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Dict, Optional

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.schema import rechunk_cube
from xcube.core.subsampling import (
    AggMethods,
    get_dataset_agg_methods,
    subsample_dataset,
)
from xcube.core.tilingscheme import get_num_levels
from xcube.util.assertions import assert_instance, assert_true

from .lazy import LazyMultiLevelDataset


class BaseMultiLevelDataset(LazyMultiLevelDataset):
    """A multi-level dataset whose level datasets are
    created by down-sampling a base dataset.

    Args:
        base_dataset: The base dataset for the level at index zero.
        grid_mapping: Optional grid mapping for *base_dataset*.
        num_levels: Optional number of levels.
        ds_id: Optional dataset identifier.
        agg_methods: Optional aggregation methods. May be given as
            string or as mapping from variable name pattern to
            aggregation method. Valid aggregation methods are None,
            "first", "min", "max", "mean", "median", and "mode".
            If None, the default, "first" is used for integer
            variables and "mean" for floating point variables.
    """

    def __init__(
        self,
        base_dataset: xr.Dataset,
        grid_mapping: Optional[GridMapping] = None,
        num_levels: Optional[int] = None,
        agg_methods: Optional[AggMethods] = "first",
        ds_id: Optional[str] = None,
    ):
        assert_instance(base_dataset, xr.Dataset, name="base_dataset")
        if grid_mapping is not None:
            assert_instance(grid_mapping, GridMapping, name="grid_mapping")

        if grid_mapping is None:
            # TODO (forman): why not computing it lazily?
            grid_mapping = GridMapping.from_dataset(base_dataset, tolerance=1e-4)

        self._agg_methods = get_dataset_agg_methods(
            base_dataset,
            xy_dim_names=grid_mapping.xy_dim_names,
            agg_methods=agg_methods,
        )

        self._base_dataset = base_dataset
        super().__init__(grid_mapping=grid_mapping, num_levels=num_levels, ds_id=ds_id)

    @property
    def agg_methods(self):
        return self._agg_methods

    def _get_num_levels_lazily(self) -> int:
        gm = self.grid_mapping
        return get_num_levels(gm.size, gm.tile_size)

    def _get_dataset_lazily(self, index: int, parameters: dict[str, Any]) -> xr.Dataset:
        """Compute the dataset at level *index*: If *index* is zero, return
        the base image passed to constructor, otherwise down-sample the
        dataset for the level at given *index*.

        Args:
            index: the level index
            parameters: currently unused

        Returns:
            the dataset for the level at *index*.
        """
        assert_instance(index, int, name="index")
        if index < 0:
            index = self.num_levels + index
        assert_true(0 <= index < self.num_levels, message="index out of range")

        if index == 0:
            level_dataset = self._base_dataset
        else:
            level_dataset = subsample_dataset(
                self._base_dataset,
                2**index,
                xy_dim_names=self.grid_mapping.xy_dim_names,
                agg_methods=self._agg_methods,
            )

        # Tile each level according to grid mapping
        tile_size = self.grid_mapping.tile_size
        if tile_size is not None:
            level_dataset, _ = rechunk_cube(
                level_dataset, self.grid_mapping, tile_size=tile_size
            )
        return level_dataset



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import unittest
from test.mixins import AlmostEqualDeepMixin


class AlmostEqualDeepMixinTest(unittest.TestCase, AlmostEqualDeepMixin):
    def test_int_and_float_7_places_default(self):
        self.assertAlmostEqualDeep(0, 0.8e-8)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(0, 0.8e-7)

    def test_int(self):
        self.assertAlmostEqualDeep(45, 45)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(45, 54)

    def test_str(self):
        self.assertAlmostEqualDeep("abc", "abc")
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep("abc", "Abc")

    def test_bool(self):
        self.assertAlmostEqualDeep(True, True)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(True, False)

    def test_set(self):
        expected = {"a", 1.1256, True}
        self.assertAlmostEqualDeep(expected, expected)
        self.assertAlmostEqualDeep(expected, {"a", 1.1256, True})
        with self.assertRaises(AssertionError):
            # We currently don't test sets
            self.assertAlmostEqualDeep(expected, {"a", 1.1251, True}, places=2)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, {"a", 1.1256, False})

    def test_dict(self):
        expected = {"a": 1.1256, "b": 5}
        self.assertAlmostEqualDeep(expected, expected)
        self.assertAlmostEqualDeep(expected, {"a": 1.1256, "b": 5})
        self.assertAlmostEqualDeep(expected, {"a": 1.1251, "b": 5}, places=3)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, {"a": 1.1251, "b": 5}, places=4)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, {"a": 1.1256, "b": 6})
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, [1, 2, 3])
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, 3456)

    def test_list(self):
        expected = ["a", 1.1256, True]
        self.assertAlmostEqualDeep(expected, expected)
        self.assertAlmostEqualDeep(expected, ["a", 1.1256, True])
        self.assertAlmostEqualDeep(expected, ("a", 1.1256, True))
        self.assertAlmostEqualDeep(expected, ["a", 1.1251, True], places=3)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, ["a", 1.1251, True], places=4)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, ["a", 1.1256, False], places=4)
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, [1, 2, 3])
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(expected, 3456)

    def test_list_dict_tuple(self):
        expected = [
            {"a": True, "b": (1.1256, 45, True)},
            {"a": False, "b": (2.1256, 46, False)},
        ]
        self.assertAlmostEqualDeep(expected, expected)
        self.assertAlmostEqualDeep(
            expected,
            [
                {"a": True, "b": (1.1256, 45, True)},
                {"a": False, "b": (2.1256, 46, False)},
            ],
        )
        self.assertAlmostEqualDeep(
            expected,
            [
                {"a": True, "b": (1.1251, 45, True)},
                {"a": False, "b": (2.1259, 46, False)},
            ],
            places=3,
        )
        with self.assertRaises(AssertionError):
            self.assertAlmostEqualDeep(
                expected,
                [
                    {"a": True, "b": (1.1251, 45, True)},
                    {"a": False, "b": (2.1259, 46, False)},
                ],
                places=4,
            )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import functools
import gzip
import math
import sys

import numpy as np
import pandas as pd
import pyproj

from xcube.core.gridmapping import CRS_CRS84
from xcube.core.select import select_subset
from xcube.core.varexpr import VarExprContext, split_var_assignment
from xcube.server.api import ApiError, ApiHandler

from ..datasets import PATH_PARAM_DATASET_ID, PATH_PARAM_VAR_NAME
from .api import api
from .config import DEFAULT_MAX_VOXEL_COUNT
from .context import VolumesContext


# noinspection PyPep8Naming
@api.route("/volumes/{datasetId}/{varName}")
class VolumesContextHandler(ApiHandler[VolumesContext]):
    @api.operation(
        operation_id="getVolume",
        summary="Get the volume data for a variable.",
        description=(
            "This is an experimental feature."
            " The volume data is returned in NRRD format."
            " Details of the NRRD format can be found at"
            " https://teem.sourceforge.net/nrrd/format.html"
        ),
        parameters=[
            PATH_PARAM_DATASET_ID,
            PATH_PARAM_VAR_NAME,
            {
                "name": "bbox",
                "in": "query",
                "description": "Bounding box in degrees (WGS-84)"
                " using format x1,y1,x2,y2",
                "schema": {
                    "type": "string",
                },
            },
            {
                "name": "startDate",
                "in": "query",
                "description": "Start date",
                "schema": {"type": "string", "format": "datetime"},
            },
            {
                "name": "endDate",
                "in": "query",
                "description": "End date",
                "schema": {"type": "string", "format": "datetime"},
            },
            {
                "name": "encoding",
                "in": "query",
                "description": "Encoding of the result",
                "schema": {
                    "type": "string",
                    "enum": ["raw", "gz"],
                    "default": "gz",
                },
            },
        ],
    )
    async def get(self, datasetId: str, varName: str):
        bbox = self.request.get_query_arg("bbox", type=str, default="")
        if bbox:
            try:
                x1, y1, x2, y2 = map(float, [p.strip() for p in bbox.split(",")])
            except (ValueError, TypeError):
                raise ApiError.BadRequest("Invalid bbox")
            bbox = x1, y1, x2, y2
        else:
            bbox = None

        start_date = self.request.get_query_arg(
            "startDate", type=pd.Timestamp, default=None
        )
        end_date = self.request.get_query_arg(
            "endDate", type=pd.Timestamp, default=None
        )

        time_range = [start_date, end_date] if start_date or end_date else None

        encoding = self.request.get_query_arg("encoding", type=str, default="gz")

        if encoding not in ("gz", "raw"):
            raise ApiError.BadRequest('Encoding must be one of "gz" or "raw"')

        ml_dataset = self.ctx.datasets_ctx.get_ml_dataset(datasetId)
        grid_mapping = ml_dataset.grid_mapping

        if bbox is not None and not grid_mapping.crs.is_geographic:
            transformer = pyproj.Transformer.from_crs(CRS_CRS84, grid_mapping.crs)
            x1, y1, x2, y2 = bbox
            (x1, x2), (y1, y2) = transformer.transform((x1, x2), (y1, y2))
            bbox = x1, y1, x2, y2

        var_name, var_expr = split_var_assignment(varName)
        if var_expr:
            dataset = self.ctx.datasets_ctx.get_dataset(datasetId).copy()
            dataset[var_name] = VarExprContext(dataset).evaluate(var_expr)
        else:
            dataset = self.ctx.datasets_ctx.get_dataset(
                datasetId, expected_var_names=[var_name]
            )

        var = select_subset(
            dataset,
            var_names=[var_name],
            time_range=time_range,
            bbox=bbox,
            grid_mapping=ml_dataset.grid_mapping,
        )[var_name]

        if var.ndim != 3:
            raise ApiError.BadRequest(f"Variable must be 3-D, got {var.ndim}-D")

        voxel_count = functools.reduce(lambda x, y: x * y, var.shape, 1)
        max_voxel_count = self.ctx.config.get("VolumesAccess", {}).get(
            "MaxVoxelCount", DEFAULT_MAX_VOXEL_COUNT
        )
        if voxel_count > max_voxel_count:
            raise ApiError.BadRequest(
                f"Volume too large, please select a smaller dataset subset."
                f" Maximum is {max_voxel_count} voxels,"
                f" got {' x '.join(map(str, var.shape))} = {voxel_count}."
            )

        # TODO (forman): allow for any dtype
        values = var.astype(dtype=np.float32).values
        if not ml_dataset.grid_mapping.is_j_axis_up:
            values = values[:, ::-1, :]
        values = np.where(np.isnan(values), 0.0, values)
        data = values.tobytes(order="C")
        if encoding == "gz":
            data = gzip.compress(data)

        block_size = 1024 * 1024
        num_blocks = math.ceil(len(data) / block_size)

        size_z, size_y, size_x = var.shape
        # TODO (forman): find more suitable normalisation
        scale_x = scale_y = 100.0 / max(size_x, size_y)
        scale_z = 100.0 / size_z

        nrrd_header = (
            "NRRD0004\n"
            "# NRRD 4 Format\n"
            "# see http://teem.sourceforge.net/nrrd/format.html\n"
            "type: float\n"  # TODO (forman): allow for any dtype
            "dimension: 3\n"
            "sizes:"
            f" {size_x} {size_y} {size_z}\n"
            f"encoding:"
            f" {encoding}\n"
            "endian:"
            f" {sys.byteorder}\n"
            "space directions:"
            f" ({scale_x},0,0) (0,{scale_y},0) (0,0,{scale_z})\n"
            "space origin: (0,0,0)\n"
            "\n"
        )

        self.response.set_header("Content-Type", "application/octet-stream")
        self.response.set_header("Cache-Control", "max-age=1")
        self.response.write(bytes(nrrd_header, "utf-8"))
        import time

        for i in range(num_blocks):
            self.response.write(data[i * block_size : i * block_size + block_size])
            time.sleep(1 / 10000)

        await self.response.finish()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import threading
from abc import ABC, abstractmethod
from asyncio import CancelledError
from concurrent.futures import Executor, Future
from concurrent.futures.thread import ThreadPoolExecutor
from enum import Enum
from typing import Any, Callable

import tabulate

from xcube.util.assertions import assert_given, assert_instance


class PreloadStatus(Enum):
    """Preload process status."""

    waiting = "waiting"
    started = "started"
    stopped = "stopped"
    cancelled = "cancelled"
    failed = "failed"

    def __str__(self):
        return self.name.upper()

    def __repr__(self):
        return f"{self.__class__.__name__}.{self.name}"


class PreloadState:
    """Preload state."""

    def __init__(
        self,
        data_id: str,
        status: PreloadStatus | None = None,
        progress: float | None = None,
        message: str | None = None,
        exception: BaseException | None = None,
    ):
        assert_given(data_id, name="data_id")
        self.data_id = data_id
        self.status = status
        self.progress = progress
        self.message = message
        self.exception = exception

    def update(self, event: "PreloadState"):
        """Update this state with the given partial state.

        Args:
            event: the partial state.
        """
        assert_instance(event, PreloadState, name="event")
        if self.data_id == event.data_id:
            if event.status is not None:
                self.status = event.status
            if event.progress is not None:
                self.progress = event.progress
            if event.message is not None:
                self.message = event.message
            if event.exception is not None:
                self.exception = event.exception

    def __str__(self):
        return ", ".join(f"{k}={v}" for k, v in _to_dict(self).items())

    def __repr__(self):
        args = ", ".join(f"{k}={v!r}" for k, v in _to_dict(self).items())
        return f"{self.__class__.__name__}({args})"


class PreloadHandle(ABC):
    """A handle for a preload job."""

    @abstractmethod
    def get_state(self, data_id: str) -> PreloadState:
        """Get the preload state for the given *data_id*.

        Args:
            data_id: The data identifier.
        Returns:
            The preload state.
        """

    @property
    @abstractmethod
    def cancelled(self) -> bool:
        """True` if the preload job has been cancelled."""

    @abstractmethod
    def cancel(self):
        """Cancel the preload job."""

    def close(self):
        """Close the preload job.

        Should be called if the preloaded data is no longer needed.

        This method usually cleans the cache associated with
        this preload object.

        The default implementation does nothing.
        """

    @abstractmethod
    def show(self) -> Any:
        """Show the current state of the preload job.

        This method is useful for non-blocking / asynchronous preload
        implementations, especially in a Jupyter Notebook context.
        In this case an implementation might want to display a widget
        suitable for in-place updating, e.g., an ``ipywidgets`` widget.
        """

    def notify(self, event: PreloadState):
        """Notify about a preload state change.

        Updates the preload job using the given partial state
        *event* that refers to state changes of a running
        preload task.

        Args:
            event: A partial state
        """

    def __enter__(self) -> "PreloadHandle":
        """Enter the context.

        Returns:
            This object.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context. Calls ``close()``."""
        self.close()


class NullPreloadHandle(PreloadHandle):
    """Null-pattern implementation of a ``PreloadHandle``."""

    def get_state(self, data_id: str) -> PreloadState:
        return PreloadState(data_id)

    @property
    def cancelled(self) -> bool:
        return False

    def cancel(self):
        pass

    def show(self) -> Any:
        return None


class ExecutorPreloadHandle(PreloadHandle):
    """An implementation of a ``PreloadHandle`` that uses a
    ``concurrent.futures.Executor`` for concurrently preloading data.

    You can either use this class

    - directly, by passing your `preload_data` function to the constructor
    - or by deriving your own class from it, then overriding the
      default implementation of its ``preload_data`` method.

    By default, the preload state updates are displayed. How, depends on the
    context: If executed in Jupyter notebooks (``IPython.display`` is available)
    the handle makes use of ``ipywidgets``, if installed, otherwise it outputs HTML.
    If not executed in a notebook it dumps process information to ``stdout``.
    If you don't want any output of preload state updates or results,
    use the ``silent`` flag.

    Args:
        data_ids: The identifiers of the data resources to be preloaded.
        preload_data: The function that preloads individual datasets.
            Optional. If not provided, you should override the
            ``preload_data`` method to implement the preloading.
        executor: Optional executor such as an instance of
            ``concurrent.futures.thread.ThreadPoolExecutor`` or
            ``concurrent.futures.process.ProcessPoolExecutor``.
            If not provided, a ``ThreadPoolExecutor`` with default settings
            will be used.
        blocking: `True` (the default) if the constructor should wait for
            all preload task to finish before the calling thread
            continues execution.
        silent: ``True`` if you don't want any preload state output.
            Defaults to ``False``.
    """

    def __init__(
        self,
        data_ids: tuple[str, ...],
        preload_data: Callable[[PreloadHandle, str], None] | None = None,
        executor: Executor | None = None,
        blocking: bool = True,
        silent: bool = False,
    ):
        self._preload_data = preload_data
        self._executor = executor or ThreadPoolExecutor()
        self._blocking = blocking

        self._states = {data_id: PreloadState(data_id=data_id) for data_id in data_ids}
        self._cancel_event = threading.Event()
        self._display = PreloadDisplay.create(list(self._states.values()), silent)
        self._silent = silent
        self._lock = threading.Lock()
        self._futures: dict[str, Future[str]] = {}
        for data_id in data_ids:
            future: Future[str] = self._executor.submit(self._run_preload_data, data_id)
            future.add_done_callback(self._handle_preload_data_done)
            self._futures[data_id] = future

        if blocking:
            if not self._silent:
                self._display.show()
            self._executor.shutdown(wait=True)

    def get_state(self, data_id: str) -> PreloadState:
        return self._states[data_id]

    @property
    def cancelled(self) -> bool:
        """Return true if and only if the internal flag is true."""
        return self._cancel_event.is_set()

    def cancel(self):
        self._cancel_event.set()
        for future in self._futures.values():
            future.cancel()
        self._executor.shutdown(wait=False)

    def notify(self, event: PreloadState):
        state = self._states[event.data_id]
        if (
            event.status is not None
            and event.status != state.status
            and state.status
            in (PreloadStatus.stopped, PreloadStatus.cancelled, PreloadStatus.failed)
        ):
            # Status cannot be changed
            return
        with self._lock:
            state.update(event)
            if not self._silent:
                self._display.update()

    def preload_data(self, data_id: str):
        """Preload the data resource given by *data_id*.

        Concurrently executes the *preload_data* passed to the constructor,
        if any. Otherwise, it does nothing.

        Can be overridden by clients to implement the actual preload operation.

        Args:
           data_id: The data identifier of the data resource to be preloaded.
        """
        if self._preload_data is not None:
            self._preload_data(self, data_id)

    def _run_preload_data(self, data_id: str) -> str:
        self.notify(PreloadState(data_id, status=PreloadStatus.started))
        self.preload_data(data_id)
        return data_id

    def _handle_preload_data_done(self, f: Future[str]):
        # Find the data_id that belongs to given feature f
        data_id: str | None = None
        for data_id, future in self._futures.items():
            if f is future:
                break
        assert data_id is not None
        try:
            _value = f.result()
            # No exceptions, notify everything seems ok
            self.notify(PreloadState(data_id, status=PreloadStatus.stopped))
        except CancelledError as e:
            # Raised if future has been cancelled
            # while executing `_run_preload_data`
            self.notify(
                PreloadState(data_id, status=PreloadStatus.cancelled, exception=e)
            )
        except Exception as e:
            # Raised if any exception occurred
            # while executing `_run_preload_data`
            self.notify(PreloadState(data_id, status=PreloadStatus.failed, exception=e))

    def show(self) -> Any:
        return self._display.show()

    def _repr_html_(self):
        return self._display.to_html()

    def __str__(self):
        return self._display.to_text()

    def __repr__(self):
        return self._display.to_text()

    def __enter__(self) -> "PreloadHandle":
        """Enter the context.

        Does nothing but returning this handle.
        Only useful when in blocking mode.

        Returns:
            This object.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context.

        Calls ``close()`` if in blocking mode.
        Otherwise, does nothing.
        """
        if self._blocking:
            self.close()


class PreloadDisplay(ABC):
    @classmethod
    def create(
        cls, states: list[PreloadState], silent: bool | None = None
    ) -> "PreloadDisplay":
        try:
            # noinspection PyUnresolvedReferences
            from IPython.display import display

            if display is not None:
                try:
                    return IPyWidgetsPreloadDisplay(states)
                except ImportError:
                    return IPyPreloadDisplay(states)
        except ImportError:
            pass
        return PreloadDisplay(states)

    def __init__(self, states: list[PreloadState]):
        self.states = states

    def _repr_html_(self) -> str:
        return self.to_html()

    def to_text(self) -> str:
        return self.tabulate(table_format="simple")

    def to_html(self) -> str:
        return self.tabulate(table_format="html")

    def tabulate(self, table_format: str = "simple") -> str:
        """Generate HTML table from job list."""
        rows = [
            [
                state.data_id,
                f"{state.status}" if state.status is not None else "-",
                (
                    f"{round(state.progress * 100)}%"
                    if state.progress is not None
                    else "-"
                ),
                state.message or "-",
                state.exception or "-",
            ]
            for state in self.states
        ]

        return tabulate.tabulate(
            rows,
            headers=["Data ID", "Status", "Progress", "Message", "Exception"],
            tablefmt=table_format,
        )

    def show(self):
        """Display the widget container."""
        print(self.to_text())

    def update(self):
        """Update the display."""
        print(self.to_text())

    def log(self, message: str):
        """Log a message to the output widget."""
        print(message)


class IPyPreloadDisplay(PreloadDisplay):
    def __init__(self, states: list[PreloadState]):
        super().__init__(states)
        from IPython import display

        self._ipy_display = display

    def show(self):
        """Display the widget container."""
        self._ipy_display.display(self.to_html())

    def update(self):
        """Update the display."""
        self._ipy_display.clear_output(wait=True)
        self._ipy_display.display(self.to_html())

    def log(self, message: str):
        """Log a message to the output widget."""
        self._ipy_display.display(message)


class IPyWidgetsPreloadDisplay(IPyPreloadDisplay):
    def __init__(self, states: list[PreloadState]):
        super().__init__(states)
        import ipywidgets

        self._state_table = ipywidgets.HTML(self.to_html())
        self._output = ipywidgets.Output()  # not used yet
        self._container = ipywidgets.VBox([self._state_table, self._output])

    def show(self):
        """Display the widget container."""
        self._ipy_display.display(self._container)

    def update(self):
        """Update the display."""
        self._state_table.value = self.to_html()

    def log(self, message: str):
        """Log a message to the output widget."""
        with self._output:
            print(message)


def _to_dict(obj: object):
    return {
        k: v
        for k, v in obj.__dict__.items()
        if isinstance(k, str) and not k.startswith("_") and v is not None
    }



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.
from collections.abc import Hashable, Mapping
from typing import Optional

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.util.assertions import assert_instance


def encode_grid_mapping(
    ds: xr.Dataset,
    gm: GridMapping,
    gm_name: Optional[str] = None,
    force: Optional[bool] = None,
) -> xr.Dataset:
    """Encode the given grid mapping *gm* into a
    copy of *ds* in a CF-compliant way and return the dataset copy.
    The function removes any existing grid mappings.

    If the CRS of *gm* is geographic and the spatial dimension and coordinate
    names are "lat", "lon" and *force* is ``False``, or *force* is ``None``
    and no former grid mapping was encoded in *ds*, then nothing else is
    done and the dataset copy is returned without further action.

    Otherwise, for every spatial data variable with dims=(..., y, x),
    the function sets the attribute "grid_mapping" to *gm_name*.
    The grid mapping CRS is encoded in a new 0-D variable named *gm_name*.

    Args:
        ds: The dataset.
        gm: The dataset's grid mapping.
        gm_name: Name for the grid mapping variable. Defaults to "crs".
        force: Whether to force encoding of grid mapping even if CRS is
            geographic and spatial dimension names are "lon", "lat".
            Optional value, if not provided, *force* will be assumed
            ``True`` if a former grid mapping was encoded in *ds*.

    Returns:
        A copy of *ds* with *gm* encoded into it.
    """
    assert_instance(ds, xr.Dataset, "ds")
    assert_instance(gm, GridMapping, "gm")
    if gm_name is not None:
        assert_instance(gm_name, str, "gm_name")

    ds_copy = ds.copy()

    x_dim_name, y_dim_name = gm.xy_dim_names
    spatial_vars = [
        (var_name, var)
        for var_name, var in ds.data_vars.items()
        if (var.ndim >= 2 and var.dims[-1] == x_dim_name and var.dims[-2] == y_dim_name)
    ]

    old_gm_names = {
        old_gm_name
        for old_gm_name in (
            var.attrs.get("grid_mapping") for var_name, var in spatial_vars
        )
        if old_gm_name and old_gm_name in ds_copy
    }
    if old_gm_names:
        force = True if force is None else force
        gm_name = gm_name or next(iter(old_gm_names))
        ds_copy = ds_copy.drop_vars(old_gm_names)

    is_geographic = (
        gm.xy_var_names == gm.xy_dim_names
        and gm.xy_dim_names == ("lon", "lat")
        and gm.crs.is_geographic
    )

    if force or not is_geographic:
        gm_name = gm_name or "crs"
        for var_name, var in spatial_vars:
            ds_copy[var_name] = var.assign_attrs(grid_mapping=gm_name)
        ds_copy[gm_name] = xr.DataArray(0, attrs=gm.crs.to_cf())

    return ds_copy


def complete_resampled_dataset(
    encode_cf: bool,
    ds: xr.Dataset,
    gm: GridMapping,
    gm_name: Optional[str],
    ref_coords: Optional[Mapping[Hashable, xr.DataArray]],
) -> xr.Dataset:
    """Internal helper."""
    if encode_cf:
        ds = encode_grid_mapping(
            ds, gm, gm_name=gm_name, force=True if gm_name else None
        )
    if ref_coords:
        compatible_coords = {
            k: v for k, v in ref_coords.items() if is_var_compatible(v, ds)
        }
        ds = ds.assign_coords(compatible_coords)
    return ds


def is_var_compatible(var: xr.DataArray, ds: xr.Dataset):
    """Internal helper."""
    for d in var.dims:
        if d in ds.sizes and var.sizes[d] != ds.sizes[d]:
            return False
    return True



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import functools
import operator
from collections.abc import Sequence
from typing import Optional, Tuple

import click

from xcube.cli.common import (
    assert_positive_int_item,
    cli_option_quiet,
    cli_option_verbosity,
    configure_cli_output,
    parse_cli_sequence,
)
from xcube.constants import FORMAT_NAME_MEM, FORMAT_NAME_NETCDF4, FORMAT_NAME_ZARR, LOG

OUTPUT_FORMAT_NAMES = [FORMAT_NAME_ZARR, FORMAT_NAME_NETCDF4, FORMAT_NAME_MEM]

DEFAULT_OUTPUT_PATH = "out.zarr"
DEFAULT_XY_NAMES = "lon,lat"
DEFAULT_DELTA = 1e-5
DEFAULT_DRY_RUN = False
DEFAULT_CRS = "EPSG:4326"


# noinspection PyShadowingBuiltins
@click.command(name="rectify", hidden=True)
@click.argument("dataset", metavar="INPUT")
@click.option(
    "--xy-vars",
    "xy_var_names",
    help=f"Comma-separated names of variables providing x any y coordinates. "
    f'If omitted, names will be guessed from available coordinate variables in INPUT, e.g. "lon,lat".',
)
@click.option(
    "--var",
    "-v",
    "var_names",
    multiple=True,
    metavar="VARIABLES",
    help="Comma-separated list of names of variables to be included or multiple options may be given. "
    "If omitted, all variables in INPUT will be reprojected.",
)
@click.option(
    "--output",
    "-o",
    "output_path",
    metavar="OUTPUT",
    default=DEFAULT_OUTPUT_PATH,
    help=f"Output path. Defaults to {DEFAULT_OUTPUT_PATH!r}.",
)
@click.option(
    "--format",
    "-f",
    "output_format",
    metavar="FORMAT",
    type=click.Choice(OUTPUT_FORMAT_NAMES),
    help="Output format. If omitted, format will be guessed from OUTPUT.",
)
@click.option(
    "--size",
    "-s",
    "output_size",
    metavar="SIZE",
    help='Output size in pixels using format "WIDTH,HEIGHT", e.g. "2048,1024". '
    "If omitted, a size will be computed so spatial resolution of INPUT is preserved.",
)
@click.option(
    "--tile_size",
    "-t",
    "output_tile_size",
    metavar="TILE_SIZE",
    help='Output tile size in pixels using format "WIDTH,HEIGHT", e.g. "512,512". '
    "If given, the output will be chunked w.r.t. TILE_SIZE. Otherwise the output will NOT be chunked.",
)
@click.option(
    "--point",
    "-p",
    "output_point",
    metavar="POINT",
    help="Output spatial coordinates of the point referring to pixel col=0.5,row=0.5 "
    'using format "LON,LAT" or "X,Y", e.g. "1.2,53.5". '
    "If omitted, the default reference point will the INPUT's minimum spatial coordinates.",
)
@click.option(
    "--res",
    "-r",
    "output_res",
    metavar="RES",
    type=float,
    help="Output spatial resolution in decimal degrees. "
    "If omitted, the default resolution will be close to the spatial resolution of INPUT.",
)
@click.option(
    "--crs",
    "-r",
    "output_crs",
    metavar="CRS",
    help="Output spatial coordinate reference system (CRS). "
    f'If omitted, the default CRS will be "{DEFAULT_CRS}".',
)
@click.option(
    "--delta",
    "-d",
    type=float,
    default=DEFAULT_DELTA,
    help="Relative maximum delta for detection whether a "
    "target pixel center is within a source pixel's boundary.",
)
@cli_option_quiet
@cli_option_verbosity
@click.option(
    "--dry-run",
    default=DEFAULT_DRY_RUN,
    is_flag=True,
    help="Just read and process INPUT, but don't produce any outputs.",
)
def rectify(
    dataset: str,
    xy_var_names: str = None,
    var_names: str = None,
    output_path: str = None,
    output_format: str = None,
    output_size: str = None,
    output_tile_size: str = None,
    output_point: str = None,
    output_res: float = None,
    output_crs: str = None,
    delta: float = DEFAULT_DELTA,
    quiet: bool = False,
    verbosity: int = 0,
    dry_run: bool = DEFAULT_DRY_RUN,
):
    """Rectify a dataset to WGS-84 using its per-pixel geo-locations."""
    configure_cli_output(quiet=quiet, verbosity=verbosity)

    input_path = dataset

    xy_var_names = parse_cli_sequence(
        xy_var_names, metavar="VARIABLES", num_items=2, item_plural_name="names"
    )
    var_name_lists = [
        parse_cli_sequence(
            var_name_specifier, metavar="VARIABLES", item_plural_name="names"
        )
        for var_name_specifier in var_names
    ]
    var_name_flat_list = functools.reduce(operator.iconcat, var_name_lists, [])

    output_size = parse_cli_sequence(
        output_size,
        metavar="SIZE",
        num_items=2,
        item_plural_name="sizes",
        item_parser=int,
        item_validator=assert_positive_int_item,
    )
    output_tile_size = parse_cli_sequence(
        output_tile_size,
        metavar="TILE_SIZE",
        num_items=2,
        item_plural_name="tile sizes",
        item_parser=int,
        item_validator=assert_positive_int_item,
    )
    output_point = parse_cli_sequence(
        output_point,
        metavar="POINT",
        num_items=2,
        item_plural_name="coordinates",
        item_parser=float,
    )

    # noinspection PyBroadException
    _rectify(
        input_path,
        xy_var_names,
        None if len(var_name_flat_list) == 0 else var_name_flat_list,
        output_path,
        output_format,
        output_size,
        output_tile_size,
        output_point,
        output_res,
        output_crs,
        delta,
        dry_run=dry_run,
        monitor=LOG.info,
    )

    return 0


def _rectify(
    input_path: str,
    xy_names: Optional[tuple[str, str]],
    var_names: Optional[Sequence[str]],
    output_path: str,
    output_format: Optional[str],
    output_size: Optional[tuple[int, int]],
    output_tile_size: Optional[tuple[int, int]],
    output_point: Optional[tuple[float, float]],
    output_res: Optional[float],
    output_crs: Optional[str],
    delta: float,
    dry_run: bool,
    monitor,
):
    import pyproj.crs

    from xcube.core.dsio import guess_dataset_format, open_dataset, write_dataset
    from xcube.core.gridmapping import GridMapping
    from xcube.core.resampling import rectify_dataset
    from xcube.core.sentinel3 import is_sentinel3_product, open_sentinel3_product

    if not output_format:
        output_format = guess_dataset_format(output_path)

    output_gm = None
    output_gm_given = (
        output_size is not None,
        output_point is not None,
        output_res is not None,
        output_crs is not None,
    )
    if all(output_gm_given):
        output_gm = GridMapping.regular(
            size=output_size,
            xy_min=output_point,
            xy_res=output_res,
            crs=pyproj.crs.CRS.from_user_input(output_crs),
        )
    elif any(output_gm_given):
        raise click.ClickException(
            "SIZE, POINT, RES, and CRS must all be given or none of them."
        )

    monitor(f"Opening dataset from {input_path!r}...")

    if is_sentinel3_product(input_path):
        src_ds = open_sentinel3_product(input_path)
    else:
        src_ds = open_dataset(input_path)

    monitor("Rectifying...")
    rectified_ds = rectify_dataset(
        src_ds,
        xy_var_names=xy_names,
        var_names=var_names,
        target_gm=output_gm,
        tile_size=output_tile_size,
        uv_delta=delta,
    )

    if rectified_ds is None:
        monitor(
            f"Dataset {input_path} does not seem to have an intersection with bounding box"
        )
        return

    monitor(f"Writing rectified dataset to {output_path!r}...")
    if not dry_run:
        write_dataset(rectified_ds, output_path, output_format)

    monitor(f"Done.")



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import JsonIntegerSchema, JsonObjectSchema

COMPUTE_CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        MaxWorkers=JsonIntegerSchema(minimum=1),
        # Executor=JsonObjectSchema(),
        # OpRegistry=JsonStringSchema(),
    ),
    additional_properties=False,
)

CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        Compute=COMPUTE_CONFIG_SCHEMA,
    )
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from chartlets import Contribution


class Panel(Contribution):
    """Panel contribution.

    A panel is a UI-contribution to xcube Viewer.
    To become effective, instances of this class must be added
    to a ``chartlets.Extension`` instance exported from your extension
    module.

    Args:
        name: A name that is unique within the extension.
        title: An initial title for the panel.
        icon: Name of a [Material Design icon](https://fonts.google.com/icons)
            to be used for the icon button representing the panel in the
            viewer's sidebar.
        position: If given, place the panel's icon button at the given position
            in the viewer's sidebar.
    """

    def __init__(
        self,
        name: str,
        title: str | None = None,
        icon: str | None = None,
        position: int | None = None,
    ):
        super().__init__(
            name,
            visible=False,
            title=title,
            icon=icon,
            position=position,
        )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import ApiError, ApiHandler

from ...datasets import (
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    QUERY_PARAM_CMAP,
    QUERY_PARAM_CRS,
    QUERY_PARAM_NORM,
    QUERY_PARAM_VMAX,
    QUERY_PARAM_VMIN,
)
from ...tiles import (
    PATH_PARAM_X,
    PATH_PARAM_Y,
    PATH_PARAM_Z,
    QUERY_PARAM_RETINA,
    QUERY_PARAM_TIME,
)
from ...tiles.controllers import compute_ml_dataset_tile
from .api import api
from .context import WmtsContext
from .controllers import (
    WMTS_CRS84_TMS_ID,
    WMTS_TILE_FORMAT,
    WMTS_VERSION,
    WMTS_WEB_MERCATOR_TMS_ID,
    get_crs_name_from_tms_id,
    get_wmts_capabilities_xml,
)

_VALID_WMTS_TMS_IDS = (WMTS_CRS84_TMS_ID, WMTS_WEB_MERCATOR_TMS_ID)

PATH_PARAM_TMS_ID = {
    "name": "tmsId",
    "in": "query",
    "description": "Tile matrix set identifier",
    "schema": {
        "type": "string",
        "enum": [WMTS_CRS84_TMS_ID, WMTS_WEB_MERCATOR_TMS_ID],
    },
}

TILE_PARAMETERS = [
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    PATH_PARAM_Z,
    PATH_PARAM_Y,
    PATH_PARAM_X,
    QUERY_PARAM_CRS,
    QUERY_PARAM_VMIN,
    QUERY_PARAM_VMAX,
    QUERY_PARAM_CMAP,
    QUERY_PARAM_NORM,
    QUERY_PARAM_TIME,
    QUERY_PARAM_RETINA,
]

TMS_TILE_PARAMETERS = [
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    PATH_PARAM_Z,
    PATH_PARAM_Y,
    PATH_PARAM_X,
    QUERY_PARAM_CRS,
    QUERY_PARAM_NORM,
    QUERY_PARAM_VMIN,
    QUERY_PARAM_VMAX,
    QUERY_PARAM_CMAP,
    QUERY_PARAM_TIME,
    QUERY_PARAM_RETINA,
]


@api.route("/wmts/1.0.0/WMTSCapabilities.xml")
class WmtsCapabilitiesXmlHandler(ApiHandler[WmtsContext]):
    @api.operation(
        operationId="getWmtsCapabilities",
        summary="Gets the WMTS capabilities as XML document",
    )
    async def get(self):
        self.request.make_query_lower_case()
        capabilities = await self.ctx.run_in_executor(
            None,
            get_wmts_capabilities_xml,
            self.ctx,
            self.request.reverse_base_url,
            WMTS_CRS84_TMS_ID,
        )
        self.response.set_header("Content-Type", "application/xml")
        await self.response.finish(capabilities)


@api.route("/wmts/1.0.0/{tmsId}/WMTSCapabilities.xml")
class WmtsCapabilitiesXmlForTmsHandler(ApiHandler[WmtsContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operationId="getWmtsTmsCapabilities",
        summary="Gets the WMTS capabilities for tile matrix set as XML document",
        parameters=[PATH_PARAM_TMS_ID],
    )
    async def get(self, tmsId):
        self.request.make_query_lower_case()
        _assert_valid_tms_id(tmsId)
        capabilities = await self.ctx.run_in_executor(
            None,
            get_wmts_capabilities_xml,
            self.ctx,
            self.request.reverse_base_url,
            tmsId,
        )
        self.response.set_header("Content-Type", "application/xml")
        await self.response.finish(capabilities)


@api.route("/wmts/1.0.0/tile/{datasetId}/{varName}/{z}/{y}/{x}.png")
class WmtsImageTileHandler(ApiHandler[WmtsContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operationId="getWmtsImageTile",
        summary="Gets a WMTS image tile in PNG format.",
        parameters=TILE_PARAMETERS,
    )
    async def get(self, datasetId: str, varName: str, z: str, y: str, x: str):
        self.request.make_query_lower_case()
        tms_id = self.request.get_query_arg("tilematrixset", default=WMTS_CRS84_TMS_ID)
        _assert_valid_tms_id(tms_id)
        crs_name = get_crs_name_from_tms_id(tms_id)
        tile = await self.ctx.run_in_executor(
            None,
            compute_ml_dataset_tile,
            self.ctx.tiles_ctx,
            datasetId,
            varName,
            crs_name,
            x,
            y,
            z,
            _query_to_dict(self.request),
        )
        self.response.set_header("Content-Type", "image/png")
        await self.response.finish(tile)


@api.route("/wmts/1.0.0/tile/{datasetId}/{varName}/{tmsId}/{z}/{y}/{x}.png")
class WmtsImageTileForTmsHandler(ApiHandler[WmtsContext]):
    # noinspection PyPep8Naming
    @api.operation(
        operationId="getWmtsTmsImageTile",
        summary="Gets a WMTS image tile for given tile matrix set in PNG format",
        parameters=TMS_TILE_PARAMETERS,
    )
    async def get(
        self, datasetId: str, varName: str, tmsId: str, z: str, y: str, x: str
    ):
        self.request.make_query_lower_case()
        _assert_valid_tms_id(tmsId)
        crs_name = get_crs_name_from_tms_id(tmsId)
        tile = await self.ctx.run_in_executor(
            None,
            compute_ml_dataset_tile,
            self.ctx.tiles_ctx,
            datasetId,
            varName,
            crs_name,
            x,
            y,
            z,
            _query_to_dict(self.request),
        )
        self.response.set_header("Content-Type", "image/png")
        await self.response.finish(tile)


@api.route("/wmts/kvp")
class WmtsKvpHandler(ApiHandler[WmtsContext]):
    @api.operation(
        operationId="invokeWmtsMethodFromKvp",
        summary="Invokes the WMTS by key-value pairs",
        parameters=TILE_PARAMETERS,
    )
    async def get(self):
        self.request.make_query_lower_case()
        service = self.request.get_query_arg("service")
        if service != "WMTS":
            raise ApiError.BadRequest('value for "service" parameter must be "WMTS"')
        request = self.request.get_query_arg("request")
        if request == "GetCapabilities":
            wmts_version = self.request.get_query_arg("version", default=WMTS_VERSION)
            if wmts_version != WMTS_VERSION:
                raise ApiError.BadRequest(
                    f'value for "version" parameter must be "{WMTS_VERSION}"'
                )
            tms_id = self.request.get_query_arg(
                "tilematrixset", default=WMTS_CRS84_TMS_ID
            )
            _assert_valid_tms_id(tms_id)
            capabilities = await self.ctx.run_in_executor(
                None,
                get_wmts_capabilities_xml,
                self.ctx,
                self.request.reverse_base_url,
                tms_id,
            )
            self.response.set_header("Content-Type", "application/xml")
            await self.response.finish(capabilities)

        elif request == "GetTile":
            wmts_version = self.request.get_query_arg("version", default=WMTS_VERSION)
            if wmts_version != WMTS_VERSION:
                raise ApiError.BadRequest(
                    f'value for "version" parameter must be "{WMTS_VERSION}"'
                )
            layer = self.request.get_query_arg("layer")
            try:
                ds_id, var_name = layer.split(".")
            except ValueError as e:
                raise ApiError.BadRequest(
                    'value for "layer" parameter must be "<dataset>.<variable>"'
                ) from e
            # For time being, we ignore "style"
            # style = self.request.get_query_arg("style"
            mime_type = self.request.get_query_arg(
                "format", default=WMTS_TILE_FORMAT
            ).lower()
            if mime_type not in (WMTS_TILE_FORMAT, "png"):
                raise ApiError.BadRequest(
                    f'value for "format" parameter must be "{WMTS_TILE_FORMAT}"'
                )
            tms_id = self.request.get_query_arg(
                "tilematrixset", default=WMTS_CRS84_TMS_ID
            )
            _assert_valid_tms_id(tms_id)
            crs_name = get_crs_name_from_tms_id(tms_id)
            x = self.request.get_query_arg("tilecol", type=int)
            y = self.request.get_query_arg("tilerow", type=int)
            z = self.request.get_query_arg("tilematrix", type=int)
            tile = await self.ctx.run_in_executor(
                None,
                compute_ml_dataset_tile,
                self.ctx.tiles_ctx,
                ds_id,
                var_name,
                crs_name,
                x,
                y,
                z,
                _query_to_dict(self.request),
            )
            self.response.set_header("Content-Type", "image/png")
            await self.response.finish(tile)
        elif request == "GetFeatureInfo":
            raise ApiError.BadRequest(
                'request type "GetFeatureInfo" not yet implemented'
            )
        else:
            raise ApiError.BadRequest(f'invalid request type "{request}"')


def _assert_valid_tms_id(tms_id: str):
    if tms_id not in _VALID_WMTS_TMS_IDS:
        raise ApiError.BadRequest(
            f'value for "tilematrixset" parameter'
            f" must be one of {_VALID_WMTS_TMS_IDS!r}"
        )


def _query_to_dict(request):
    return {k: v[0] for k, v in request.query.items()}



#  Copyright (c) 2018-2025 by xcube team and contributors
#  Permissions are hereby granted under the terms of the MIT License:
#  https://opensource.org/licenses/MIT.

from chartlets import Extension

from .panel_histo2d import panel as histo2d_panel
from .panel_spectrum import panel as spectrum_panel
from .panel_demo import panel as demo_panel

ext = Extension(__name__)
ext.add(histo2d_panel)
ext.add(spectrum_panel)
ext.add(demo_panel)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Optional


def split_var_assignment(var_name_or_assign: str) -> tuple[str, Optional[str]]:
    """Split *var_name_or_assign* into a variable name and expression part.

    Args:
        var_name_or_assign: A variable name or an expression

    Return:
        A pair (var_name, var_expr) if *var_name_or_assign* is an assignment
        expression, otherwise (var_name, None).
    """
    if "=" in var_name_or_assign:
        var_name, var_expr = map(
            lambda s: s.strip(), var_name_or_assign.split("=", maxsplit=1)
        )
        return var_name, var_expr
    else:
        return var_name_or_assign, None



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import numpy as np
import xarray as xr


def is_empty_cube(cube: xr.Dataset) -> bool:
    return len(cube.data_vars) == 0


def strip_cube(cube: xr.Dataset) -> xr.Dataset:
    drop_vars = [
        k
        for k, v in cube.data_vars.items()
        if len(v.shape) < 3
        or np.prod(v.shape) == 0
        or v.shape[-2] < 2
        or v.shape[-1] < 2
    ]
    if drop_vars:
        return cube.drop_vars(drop_vars)
    return cube



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import functools
import logging
import time
from contextlib import AbstractContextManager
from typing import Optional

from xcube.constants import LOG


def measure_time_cm(logger=None, disabled=False):
    """Get a context manager for measuring execution time of code blocks
    and logging the result.

    Measure duration and log output:::

        measure_time = measure_time_cm()
        with measure_time("heavy computation"):
            do_heavy_computation()

    or just measure duration:::

        with measure_time() as cm:
            do_heavy_computation()
        print("heavy computation took %2.f seconds" % cm.duration)

    Args:
        logger: The logger to be used. May be a string or logger object.
            Defaults to "xcube".
        disabled: If True, efficiently disables timing and logging.

    Returns:
        a context manager callable
    """
    if disabled:
        return _do_not_measure_time_cm
    else:
        return functools.partial(measure_time, logger=logger)


class measure_time(AbstractContextManager):
    def __init__(self, *args, logger: Optional[logging.Logger] = None, **kwargs):
        self.message = args[0] if args else None
        self.args = args[1:] if args else ()
        self.kwargs = kwargs
        if isinstance(logger, str):
            self.logger = logging.getLogger(logger)
        elif logger is None:
            self.logger = LOG
        else:
            self.logger = logger
        self.start_time = None
        self.duration = None

    def __enter__(self):
        self.start_time = time.perf_counter()
        return self

    def __exit__(self, *exc):
        self.duration = time.perf_counter() - self.start_time
        if self.message:
            self.logger.info(
                self.message + ": took %.2fms",
                *self.args,
                self.duration * 1000,
                **self.kwargs,
            )


class _do_not_measure_time_cm(AbstractContextManager):
    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwargs):
        self.duration = None
        self.logger = None
        self.message = None

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        pass
